import collections
import functools
import hashlib
import itertools
import queue
import random
import string
import warnings
from typing import List, Set

try:
    from . import zdlexer

except ImportError:
    import zdlexer


actor_list = {}

_user_var_setters = {
    "int": "A_SetUserVar",
    "float": "A_SetUserVarFloat",
}

_user_array_setters = {
    "int": "A_SetUserArray",
    "float": "A_SetUserArrayFloat",
}


def stringify(content):
    if isinstance(content, str):
        if content[0] in "'\"" and content[-1] == content[0] and len(content) > 1:
            return content

        return '"' + repr(content)[1:-1] + '"'

    return repr(content)


def unstringify(content):
    if content[0] in "'\"" and content[-1] == content[0] and len(content) > 1:
        return content[1:-1]

    return content


def make_id(length=30):
    """Returns an ID, which can be stored in the ZDCode
    to allow namespace compatibility between multiple
    ZDCode mods."""
    return "".join(
        random.choice(string.ascii_letters + string.digits) for _ in range(length)
    )


def decorate(o, indent=4):
    """Get the object's compiled DECORATE code."""
    return o.to_decorate().to_string(indent)


class TextNode:
    def __init__(self, lines=(), indent=1):
        self.lines = list(lines)
        self.indent = indent

    def add_line(self, line):
        self.lines.append(line)

    def __str__(self):
        return "\n".join(
            "\n".join("\t" * self.indent + x for x in str(l).split("\n"))
            for l in self.lines
        )

    def __len__(self):
        return len(str(self))

    def __getitem__(self, ind):
        return self.lines[ind]

    def to_string(self, tab_size=4):
        return "\n".join(str(l) for l in self.lines).replace("\t", " " * tab_size)


# ZDCode Classes
class ZDProperty:
    def __init__(self, actor, name, value):
        self.actor = actor
        self.name = name.strip()
        self.value = value

        self.actor.properties.append(self)

    def to_decorate(self):
        return TextNode(["{} {}".format(self.name, self.value)])


class ZDState:
    def __init__(
        self, sprite='"####"', frame='"#"', duration=0, keywords=None, action=None
    ):
        if not keywords:
            keywords = []

        self.sprite = sprite
        self.frame = frame
        self.keywords = keywords or []
        self.action = action
        self.duration = duration

    def clone(self):
        return ZDState(
            self.sprite, self.frame, self.duration, list(self.keywords), self.action
        )

    def state_containers(self):
        return
        yield

    def num_states(self):
        return 1

    def to_decorate(self):
        if self.keywords:
            keywords = [" "] + self.keywords

        else:
            keywords = []

        action = ""
        if self.action:
            action = " " + str(self.action)

        return TextNode(
            [
                "{} {} {}{}{}".format(
                    self.sprite.upper(),
                    self.frame.upper(),
                    str(self.duration),
                    " ".join(keywords),
                    action,
                )
            ]
        )

    def __repr__(self):
        return "<ZDState({} {} {}{}{})>".format(
            self.sprite,
            self.frame,
            self.duration,
            "+" if self.keywords else "",
            " ..." if self.action else "",
        )


class ZDDummyActor:
    def __init__(self, code, context=None, _id=None):
        if context:
            self.context = context.derive()

        else:
            self.context = ZDCodeParseContext()

        ZDBaseActor.__init__(self, code, "__dummy__", None, None, None, _id)

    def get_context(self):
        return self.context

    def to_decorate(self):
        raise NotImplementedError


class ZDDummyLabel(object):
    def __init__(self, actor, states=None):
        if not states:
            states = []

        self.name = None
        self.states = states
        self._actor = actor

    def __repr__(self):
        return "[dummy state]"

    def label_name(self):
        raise NotImplementedError

    def to_decorate(self):
        raise NotImplementedError

    def add_state(self, state):
        self.states.append(state)


class ZDLabel(object):
    def __init__(self, _actor, name, states=None, auto_append=True):
        if not states:
            states = []

        self.name = name.strip()
        self.states = states
        self._actor = _actor

        if auto_append:
            self._actor.labels.append(self)

    def __repr__(self):
        return "{}({} of {})".format(
            type(self).__name__, self.name, repr(self._actor.name)
        )

    def label_name(self):
        return self.name

    def add_state(self, state):
        self.states.append(state)

    def to_decorate(self):
        if self.name.startswith("F_"):
            self.name = "_" + self.name

        r = TextNode()
        r.add_line("{}:".format(self.name))

        for s in self.states:
            r.add_line(s.to_decorate())

        return r


class ZDRawDecorate(object):
    def __init__(self, raw):
        self.raw = raw

    def state_containers(self):
        return
        yield

    def num_states(self):
        return 0

    def to_decorate(self):
        return TextNode([self.raw], 0)


class ZDBaseActor(object):
    def __init__(
        self, code, name=None, inherit=None, replace=None, doomednum=None, _id=None
    ):
        self.code = code
        self.name = name.strip()
        self.inherit = inherit
        self.replace = replace
        self.num = doomednum
        self.id = _id or make_id(30)


class ZDActor(ZDBaseActor):
    def __init__(
        self,
        code,
        name,
        inherit=None,
        replace=None,
        doomednum=None,
        _id=None,
        context=None,
    ):
        if context:
            self.context = context.derive()

        else:
            self.context = ZDCodeParseContext()

        ZDBaseActor.__init__(self, code, name, inherit, replace, doomednum, _id)

        self.labels = []
        self.properties = []
        self.flags = set()
        self.uservars = []
        self.antiflags = set()
        self.funcs = []
        self.namefuncs = {}
        self.raw = []
        self.doomednum = doomednum

        if self.inherit and self.inherit.upper() in code.actor_names:
            self.all_funcs = list(code.actor_names[self.inherit.upper()].all_funcs)
            self.context.update(code.actor_names[self.inherit.upper()].context)

        else:
            self.all_funcs = []

        # code.actor_names[name.upper()] = self

    def __repr__(self):
        return "<ZDActor({}{}{}{})>".format(
            self.name,
            self.inherit and ("extends " + self.inherit) or "",
            self.replace and ("replaces " + self.replace or "") or "",
            self.doomednum and ("#" + str(self.doomednum)) or "",
        )

    def get_context(self):
        return self.context

    def make_spawn_label(self):
        return ZDLabel(
            self,
            "Spawn",
            [ZDRawDecorate("goto Super::Spawn" if self.inherit else "stop")],
            auto_append=False,
        )

    def get_spawn_label(self):
        for l in self.labels:
            if l.name.upper() == "SPAWN":
                return l

        return None

    def _set_user_var_state(self, var):
        vtype, vlit = var["value"]

        if vtype == "val":
            return [
                ZDState(
                    action="{}({}, {})".format(
                        _user_var_setters[var["type"]], stringify(var["name"]), vlit
                    )
                )
            ]

        elif vtype == "arr":
            return [
                ZDState(
                    action="{}({}, {}, {})".format(
                        _user_array_setters[var["type"]], stringify(var["name"]), i, v
                    )
                )
                for i, v in enumerate(vlit)
            ]

    def _get_spawn_prelude(self):
        return sum(
            [
                self._set_user_var_state(var)
                for var in self.uservars
                if var.get("value", None)
            ],
            [],
        )

    def prepare_spawn_label(self):
        label = self.get_spawn_label()

        if self.uservars:
            if not label:
                label = self.make_spawn_label()

            label.states = (
                [ZDState("TNT1", "A")] + self._get_spawn_prelude() + label.states
            )

        elif label:
            label.states.insert(0, ZDState("TNT1", "A"))

    def top(self):
        r = TextNode()

        for p in sorted(self.properties, key=lambda p: p.name):
            r.add_line(decorate(p))

        r.add_line("")

        for u in self.uservars:
            r.add_line(
                "var {} {}{};".format(
                    u["type"],
                    u["name"],
                    "[{}]".format(u["size"]) if u.get("size", 0) else "",
                )
            )

        for f in self.flags:
            r.add_line("+{}".format(f))

        for a in self.antiflags:
            r.add_line("-{}".format(a))

        for rd in self.raw:
            r.add_line(rd)

        if len(r) == 1 and r[0].strip() == "":
            return "    "

        return r

    def label_code(self):
        r = TextNode()

        for f in self.funcs:
            r.add_line(decorate(f[1]))

        for l in self.labels:
            r.add_line(decorate(l))

        return r

    def header(self):
        r = self.name

        if self.inherit:
            r += " : {}".format(self.inherit)
        if self.replace:
            r += " replaces {}".format(self.replace)
        if self.num:
            r += " {}".format(str(self.num))

        return r

    def to_decorate(self):
        if self.labels + self.funcs:
            return TextNode(
                [
                    "Actor " + self.header(),
                    "{",
                    self.top(),
                    TextNode(
                        [
                            "States {",
                            self.label_code(),
                            "}",
                        ]
                    ),
                    "}",
                ],
                0,
            )

        return TextNode(["Actor " + self.header(), "{", self.top(), "}"], 0)


class ZDClassTemplate(ZDBaseActor):
    def __init__(
        self,
        template_parameters,
        parse_data,
        abstract_label_names,
        abstract_macro_names,
        abstract_array_names,
        group_name,
        code,
        name,
        inherit=None,
        replace=None,
        doomednum=None,
        _id=None,
    ):
        ZDBaseActor.__init__(self, code, name, inherit, replace, doomednum, _id)

        self.group_name = group_name or None

        self.template_parameters = list(template_parameters)

        self.abstract_label_names = set(abstract_label_names)
        self.abstract_macro_names = dict(abstract_macro_names)
        self.abstract_array_names = dict(abstract_array_names)

        self.parse_data = parse_data
        self.parametric_table = {}

    def generated_class_name(self, parameter_values, new_id):
        hash = hashlib.sha1()

        hash.update(self.name.encode("utf-8"))
        hash.update(self.id.encode("utf-8"))
        hash.update(new_id.encode("utf-8"))

        for parm in parameter_values:
            hash.update(parm.encode("utf-8"))

        for name in self.abstract_label_names:
            hash.update(name.encode("utf-8"))

        for name, args in self.abstract_macro_names.items():
            hash.update(name.encode("utf-8"))

            for arg in args:
                hash.update(arg.encode("utf-8"))

        return "{}__deriv_{}".format(self.name, hash.hexdigest())

    def generate_init_class(
        self,
        code,
        context,
        parameter_values,
        provided_label_names=(),
        provided_macro_names=(),
        provided_array_names=(),
        name=None,
        pending=None,
    ):
        if (
            tuple(parameter_values) in self.parametric_table
            and not self.abstract_label_names
            and not self.abstract_macro_names
            and not self.abstract_array_names
        ):
            return (False, self.parametric_table[tuple(parameter_values)])

        provided_label_names = set(provided_label_names)
        provided_macro_names = dict(provided_macro_names)

        new_name = (
            name
            if name is not None
            else self.generated_class_name(parameter_values, make_id(40))
        )

        if self.group_name:
            self.code.groups[self.group_name].append(stringify(new_name))

        context_new = context.derive("derivation of template {}".format(self.name))

        context_new.replacements.update(self.get_init_replacements(parameter_values))

        inh = (
            self.inherit
            and context_new.replacements.get(self.inherit.upper(), self.inherit)
            or None
        )
        rep = (
            self.replace
            and context_new.replacements.get(self.replace.upper(), self.replace)
            or None
        )

        context_new.replacements["SELF"] = stringify(new_name)

        res = ZDActor(self.code, new_name, inh, rep, self.num, context=context_new)

        for l in self.abstract_label_names:
            if l not in provided_label_names:
                raise CompilerError(
                    "Tried to derive template {} in {}, but abstract label {} does not have a definition!".format(
                        self.name, context.describe(), l
                    )
                )

        for m, a in self.abstract_macro_names.items():
            if m not in provided_macro_names.keys():
                raise CompilerError(
                    "Tried to derive template {} in {}, but abstract macro {} does not have a definition!".format(
                        self.name, context.describe(), m
                    )
                )

            if len(a) != len(provided_macro_names[m]):
                raise CompilerError(
                    "Tried to derive template {} in {}, but abstract macro {} has the wrong number of arguments: expected {}, got {}!".format(
                        self.name,
                        context.describe(),
                        m,
                        len(a),
                        len(provided_macro_names[m]),
                    )
                )

        for m, a in self.abstract_array_names.items():
            if m not in provided_array_names.keys():
                raise CompilerError(
                    "Tried to derive template {} in {}, but abstract array {} is not defined!".format(
                        self.name, context.describe(), m
                    )
                )

            if a["size"] != "any" and a["size"] != provided_array_names[m]:
                raise CompilerError(
                    "Tried to derive template {} in {}, but abstract array {} has a size constraint; expected {} array elements, got {}!".format(
                        self.name,
                        context.describe(),
                        m,
                        a["size"],
                        provided_array_names[m],
                    )
                )

        self.parametric_table[tuple(parameter_values)] = res

        return (True, res)

    def get_init_replacements(self, parameter_values):
        return dict(
            zip((p.upper() for p in self.template_parameters), parameter_values)
        )


class ZDBlock:
    def __init__(self, actor, states=()):
        self._actor = actor
        self.states = list(states)

    def clone(self):
        return ZDBlock(self._actor, (s.clone() for s in self.states))

    def num_states(self):
        return sum(x.num_states() for x in self.states)

    def state_containers(self):
        yield self.states

    def to_decorate(self):
        return TextNode(x.to_decorate() for x in self.states)


class ZDIfStatement(object):
    def __init__(self, actor, condition, states=()):
        self._actor = actor
        self.true_condition = condition
        self.states = list(states)
        self.else_block = None

    def clone(self):
        res = ZDIfStatement(
            self._actor, self.true_condition, (s.clone() for s in self.states)
        )
        res.set_else(self.else_block.clone())

        return res

    def state_containers(self):
        yield self.states

        if self.else_block:
            yield from self.else_block.state_containers()

    def set_else(self, else_block):
        self.else_block = else_block

    def num_block_states(self):
        return sum(x.num_states() for x in self.states)

    def num_else_states(self):
        return self.else_block.num_states()

    def num_states(self):
        if self.else_block:
            return self.num_block_states() + self.num_else_states() + 3

        else:
            return self.num_block_states() + 2

    def to_decorate(self):
        num_st_bl = self.num_block_states()

        if self.else_block:
            num_st_el = self.num_else_states()

            return TextNode(
                [
                    "TNT1 A 0 A_JumpIf({}, {})".format(
                        self.true_condition, num_st_el + 2
                    ),
                    self.else_block.to_decorate(),
                    "TNT1 A 0 A_Jump(256, {})".format(num_st_bl + 1),
                    TextNode([x.to_decorate() for x in self.states]),
                    "TNT1 A 0",
                ]
            )

        else:
            return TextNode(
                [
                    "TNT1 A 0 A_JumpIf(!({}), {})\n".format(
                        self.true_condition, num_st_bl + 1
                    ),
                    TextNode([x.to_decorate() for x in self.states]),
                    "TNT1 A 0",
                ]
            )


class ZDIfJumpStatement(object):
    def __init__(self, actor, condition_gen, states=()):
        self._actor = actor
        self.true_condition = condition_gen
        self.states = list(states)
        self.else_block = None

    def clone(self):
        res = ZDIfJumpStatement(
            self._actor, self.true_condition, (s.clone() for s in self.states)
        )
        res.set_else(self.else_block.clone())

        return res

    def state_containers(self):
        yield self.states

        if self.else_block:
            yield from self.else_block.state_containers()

    def set_else(self, else_block):
        self.else_block = else_block

    @classmethod
    def generate(cls, actor, states=()):
        def _decorator(condition_gen):
            return cls(actor, condition_gen, states)

        return _decorator

    def num_block_states(self):
        return sum(x.num_states() for x in self.states)

    def num_else_states(self):
        return self.else_block.num_states()

    def num_states(self):
        if self.else_block:
            return self.num_block_states() + self.num_else_states() + 3

        else:
            return self.num_block_states() + 3

    def to_decorate(self):
        num_st_bl = self.num_block_states()

        if self.else_block:
            num_st_el = self.num_else_states()

            return TextNode(
                [
                    "TNT1 A 0 {}".format(self.true_condition(num_st_el + 2)),
                    self.else_block.to_decorate(),
                    "TNT1 A 0 A_Jump(256, {})\n".format(num_st_bl + 1),
                    TextNode([x.to_decorate() for x in self.states]),
                    "TNT1 A 0",
                ]
            )

        else:
            return TextNode(
                [
                    "TNT1 A 0 {}".format(self.true_condition(2)),
                    "TNT1 A 0 A_Jump(256, {})\n".format(num_st_bl + 1),
                    TextNode([x.to_decorate() for x in self.states]),
                    "TNT1 A 0",
                ]
            )


class ZDSometimes(object):
    def __init__(self, actor, chance, states):
        self._actor = actor
        self.chance = chance
        self.states = states

    def clone(self):
        return ZDSometimes(self._actor, self.chance, (s.clone() for s in self.states))

    def num_states(self):
        return sum(x.num_states() for x in self.states) + 2

    def state_containers(self):
        yield self.states

    def to_decorate(self):
        num_st = sum(x.num_states() for x in self.states)

        res = TextNode()
        res.add_line(
            "TNT1 A 0 A_Jump(256-(256*({})/100), {})".format(self.chance, num_st + 1)
        )

        for x in self.states:
            res.add_line(x.to_decorate())

        res.add_line("TNT1 A 0")

        return res


num_whiles = 0


class ZDWhileStatement(object):
    def __init__(self, actor, condition, states=()):
        self._actor = actor
        self.true_condition = condition
        self.states = list(states)
        self.else_block = None

        global num_whiles

        self._while_id = num_whiles
        num_whiles += 1

        self._loop_id = "_loop_while_" + str(self._while_id)

    def clone(self):
        res = ZDWhileStatement(
            self._actor, self.true_condition, (s.clone() for s in self.states)
        )
        res.set_else(self.else_block.clone())

        return res

    def state_containers(self):
        yield self.states

        if self.else_block:
            yield from self.else_block.state_containers()

    def set_else(self, else_block):
        self.else_block = else_block

    def num_block_states(self):
        return sum(x.num_states() for x in self.states)

    def num_else_states(self):
        return self.else_block.num_states()

    def num_states(self):
        if self.else_block:
            return self.num_block_states() + self.num_else_states() + 4

        else:
            return self.num_block_states() + 3

    def to_decorate(self):
        num_st_bl = self.num_block_states()

        if self.else_block:
            num_st_el = self.num_else_states()

            return TextNode(
                [
                    "TNT1 A 0 A_JumpIf({}, {})".format(
                        self.true_condition, num_st_el + 2
                    ),
                    self.else_block.to_decorate(),
                    "TNT1 A 0 A_Jump(256, {})".format(num_st_bl + 2),
                    "{}:".format(self._loop_id),
                    TextNode([x.to_decorate() for x in self.states]),
                    "TNT1 A 0 A_JumpIf({}, {})".format(
                        self.true_condition, stringify(self._loop_id)
                    ),
                    "TNT1 A 0",
                ]
            )

        else:
            return TextNode(
                [
                    "TNT1 A 0 A_JumpIf(!({}), {})".format(
                        self.true_condition, num_st_bl + 2
                    ),
                    "{}:".format(self._loop_id),
                    TextNode([x.to_decorate() for x in self.states]),
                    "TNT1 A 0 A_JumpIf({}, {})".format(
                        self.true_condition, stringify(self._loop_id)
                    ),
                    "TNT1 A 0",
                ]
            )


class ZDWhileJumpStatement(object):
    def __init__(self, actor, condition_gen, states=()):
        self._actor = actor
        self.true_condition = condition_gen
        self.states = list(states)
        self.else_block = None

        global num_whiles

        self._while_id = num_whiles
        num_whiles += 1

        self._loop_id = "_loop_while_" + str(self._while_id)

    def clone(self):
        res = ZDWhileJumpStatement(
            self._actor, self.true_condition, (s.clone() for s in self.states)
        )
        res.set_else(self.else_block.clone())

        return res

    def state_containers(self):
        yield self.states

        if self.else_block:
            yield from self.else_block.state_containers()

    def set_else(self, else_block):
        self.else_block = else_block

    @classmethod
    def generate(cls, actor, states=()):
        def _decorator(condition_gen):
            return cls(actor, condition_gen, states)

        return _decorator

    def num_block_states(self):
        return sum(x.num_states() for x in self.states)

    def num_else_states(self):
        return self.else_block.num_states()

    def num_states(self):
        if self.else_block:
            return self.num_block_states() + self.num_else_states() + 4

        else:
            return self.num_block_states() + 4

    def to_decorate(self):
        num_st_bl = self.num_block_states()

        if self.else_block:
            num_st_el = self.num_else_states()

            return TextNode(
                [
                    "TNT1 A 0 {}".format(self.true_condition(num_st_el, 2)),
                    self.else_block.to_decorate(),
                    "TNT1 A 0 A_Jump(256, {})".format(num_st_bl + 2),
                    "{}:".format(self._loop_id),
                    TextNode([x.to_decorate() for x in self.states]),
                    "TNT1 A 0 {}".format(self.true_condition(stringify(self._loop_id))),
                    "TNT1 A 0",
                ]
            )

        else:
            return TextNode(
                [
                    "TNT1 A 0 {}".format(self.true_condition(2)),
                    "TNT1 A 0 A_Jump(256, {})".format(num_st_bl + 2),
                    "{}:".format(self._loop_id),
                    TextNode([x.to_decorate() for x in self.states]),
                    "TNT1 A 0 {}".format(self.true_condition(stringify(self._loop_id))),
                    "TNT1 A 0",
                ]
            )


class ZDInventory(object):
    def __init__(self, code, name):
        self.name = name.strip()
        self.code = code

        code.inventories.append(self)

    def to_decorate(self):
        return TextNode(
            ["Actor {} : Inventory {{Inventory.MaxAmount 1}}".format(self.name)]
        )


class ZDSkip:
    def __init__(self, code, skip_context, curr_ind):
        self.code = code
        self.context = skip_context
        self.ind = curr_ind

    def clone(self):
        raise NotImplementedError(
            "State group skipping (e.g. return in macros) could not be cloned. Injected macros cannot be cloned - please don't inject them until all else is resolved!"
        )

    def state_containers(self):
        return
        yield

    def to_decorate(self):
        return "TNT1 A 0 A_Jump(256, {})".format(
            self.context.num_states() - self.ind + 1
        )

    def num_states(self):
        return 1


# Compiler!
class CompilerError(Exception):
    pass


class ZDCodeParseContext(object):
    def __init__(
        self,
        replacements=(),
        macros=(),
        templates=(),
        actors=(),
        mods=(),
        applied_mods=None,
        remote_offset=0,
    ):
        self.macros = collections.ChainMap({}, macros)
        self.replacements = collections.ChainMap({}, replacements)
        self.templates = collections.ChainMap({}, templates)
        self.mods = collections.ChainMap({}, mods)

        self.always_applied_mods = applied_mods
        self.applied_mods = []

        self.actor_lists = list(actors) if actors else [[]]
        self.desc_stack = []
        self.states = []
        self.remote_children = []
        self.remote_offset = remote_offset

    def get_applied_mods(self):
        if self.always_applied_mods is None:
            return iter(self.applied_mods)

        else:
            return itertools.chain(self.always_applied_mods, self.applied_mods)

    def print_state_tree(self, _print_low=print, _print=print, prefix="+ "):
        ended = 0

        _print_low = _print_low or print

        def _print_top(name):
            _print_low(prefix + name)

        def _branch(l="", end=False):
            if end:
                _print("'---+ " + l)

            else:
                _print("+---+ " + l)

        def _print_next(line=""):
            if ended == 0:
                _print("|   " + line)

            elif ended == 1 and self.remote_children:
                _print(":   " + line)

            else:
                _print("    " + line)

        _print_top(
            "{} ({}/{})".format(
                self.desc_stack[-1], self.num_states(), self.remote_num_states()
            )
        )

        if self.states:
            _print("|")

            imax = len(self.states)

            for i, s in enumerate(self.states):
                ended = 1 if i >= imax - 1 else 0

                if isinstance(s, ZDCodeParseContext):
                    s.print_state_tree(
                        _print, _print_next, "'---+ " if ended > 0 else "+---+ "
                    )

                else:
                    _branch("{} ({})".format(type(s).__name__, s.num_states()), ended)

                if not ended:
                    _print("|")

        if self.remote_children:
            _print("&")
            _print(":")

            imax = len(self.remote_children)

            for i, ch in enumerate(self.remote_children):
                ended = 2 if i >= imax - 1 else 1

                ch.print_state_tree(
                    _print,
                    _print_next,
                    "^---* (remote) " if ended > 1 else "%---* (remote) ",
                )

                if not ended:
                    _print(":")

    def num_states(self):
        return sum(s.num_states() for s in self.states)

    def remote_num_states(self):
        return (
            self.remote_offset
            + sum(
                s.remote_num_states()
                if isinstance(s, ZDCodeParseContext)
                else s.num_states()
                for s in self.states
            )
            + sum(c.remote_num_states() for c in self.remote_children)
        )

    def remote_derive(
        self, desc: str = None, remote_offset: int = 0
    ) -> "ZDCodeParseContext":
        # derives without adding to states
        res = ZDCodeParseContext(
            self.replacements,
            self.macros,
            self.templates,
            self.actor_lists,
            self.mods,
            self.applied_mods,
            remote_offset,
        )
        res.desc_stack = list(self.desc_stack)

        if desc:
            res.desc_stack.append(desc)

        self.remote_children.append(res)

        return res

    def derive(self, desc: str = None) -> "ZDCodeParseContext":
        res = ZDCodeParseContext(
            self.replacements,
            self.macros,
            self.templates,
            self.actor_lists,
            self.mods,
            self.applied_mods,
        )
        res.desc_stack = list(self.desc_stack)

        if desc:
            res.desc_stack.append(desc)

        self.states.append(res)

        return res

    def __repr__(self):
        return "ZDCodeParseContext({})".format(self.repr_describe())

    def desc_block(self, desc: str):
        return ZDCtxDescBlock(self, desc)

    def update(self, other_ctx: "ZDCodeParseContext"):
        self.macros.update(other_ctx.macros)
        self.replacements.update(other_ctx.replacements)
        self.templates.update(other_ctx.templates)
        self.mods.update(other_ctx.mods)

    def add_actor(self, ac: ZDActor):
        for al in self.actor_lists:
            al.append(ac)

    def describe(self):
        return " at ".join(self.desc_stack[::-1])

    def repr_describe(self):
        return ", ".join(self.desc_stack[::-1])

    def resolve(self, name, desc="a parametrizable name"):
        while name[0] == "@":
            resolves = len(name) - len(name.lstrip("@"))
            casename = name[resolves:]
            name = name[resolves:].upper()

            if name in self.replacements:
                if resolves > 1:
                    name = "@" * (resolves - 1) + self.replacements[name]

                else:
                    name = self.replacements[name]

            else:
                raise CompilerError(
                    "No such replacement {} while trying to resolve {} in {}!".format(
                        repr(name), repr(casename), self.describe()
                    )
                )

        return name


@functools.total_ordering
class PendingTask:
    def __init__(self, priority, func):
        self.priority = priority
        self.func = func

    def __lt__(self, other):
        return self.priority < other.priority

    def __eq__(self, other):
        return self.priority == other.priority


def pending_task(priority):
    def _decorator(func):
        return PendingTask(priority, func)

    return _decorator


class ZDCtxDescBlock:
    def __init__(self, ctx, desc):
        self.ctx = ctx
        self.desc = desc

    def __enter__(self):
        self.ctx.desc_stack.append(self.desc)

    def __exit__(self, _1, _2, _3):
        assert self.ctx.desc_stack.pop() == self.desc


class ZDModClause:
    def __init__(self, code: "ZDCode", ctx: ZDCodeParseContext, selector, effects):
        self.code = code
        self.context = ctx
        self.selector = selector
        self.effects = effects

    def apply(self, ctx: ZDCodeParseContext, target_states):
        clause_ctx = self.context.derive("mod clause")
        clause_ctx.update(ctx)

        res = []

        for s in target_states:
            if self.selector(self.code, clause_ctx, s):
                alist = [s]

                for eff in self.effects:
                    nlist = []

                    for a in alist:
                        l = list(eff(self.code, clause_ctx, a))
                        nlist.extend(l)

                    alist = nlist

                res.extend(alist)

            else:
                for container in s.state_containers():
                    self.apply(ctx, container)

                res.append(s)

        target_states.clear()
        target_states.extend(res)


class ZDCode:
    class ZDCodeError(BaseException):
        pass

    @classmethod
    def parse(cls, code, fname=None, dirname=".", error_handler=None, preproc_defs=()):
        data = zdlexer.parse_code(
            code.strip(" \t\n"),
            dirname=dirname,
            filename=fname,
            error_handler=error_handler,
            preproc_defs=preproc_defs,
        )

        if data:
            res = cls()

            try:
                res._parse(data)

            except CompilerError as err:
                if error_handler:
                    error_handler(err)
                return None

            return res

        else:
            return None

    def add(self, code, fname=None, dirname=".", error_handler=None, preproc_defs=()):
        data = zdlexer.parse_code(
            code.strip(" \t\n"),
            dirname=dirname,
            filename=fname,
            error_handler=error_handler,
            imports=self.includes,
            preproc_defs=preproc_defs,
        )

        if data:
            try:
                self._parse(data)

            except CompilerError as err:
                if error_handler:
                    error_handler(err)
                return False

        return bool(data)

    def _parse_operation(self, operation, context):
        ooper, oargs = operation

        oargs = (self._parse_evaluation(a, context) for a in oargs)

        return ooper(*oargs)

    def _parse_evaluation(self, evaluation, context):
        etype, econt = evaluation

        if etype == "literal":
            return econt

        if etype == "operation":
            return self._parse_operation(econt, context)

    def _parse_expression(self, expr, context):
        etype, exval = expr

        if etype == "expr":
            return " ".join(self._parse_expression(item, context) for item in exval)

        if etype == "literal":
            return self._parse_literal(exval, context)

        if etype == "array index":
            return "[" + self._parse_expression(exval, context) + "]"

        if etype == "oper":
            return exval

        if etype == "paren expr":
            return "(" + self._parse_expression(exval, context) + ")"

    def _parse_argument(self, arg, context, name=None):
        atype, aval = arg

        if atype == "position arg":
            return self._parse_parameter(aval, context, name)

    def _parse_parameter(self, parm, context, name=None):
        ptype, pval = parm

        if ptype == "expression":
            return self._parse_expression(pval, context)

        if ptype == "template derivation":
            return self._parse_template_derivation(pval, context)

        if ptype == "anonymous class":
            return self._parse_anonym_class(pval, context)

        if ptype == "anonymous macro":
            return self._parse_anonym_macro(*pval, context, name)

    def _parse_anonym_macro(self, args, body, context, name=None):
        name = name or "ANONYMMACRO_{}_{}".format(
            self.id.upper(), self.num_anonym_macros
        )
        self.num_anonym_macros += 1

        context.macros[name.upper()] = (args, body)

        return stringify(name)

    def _parse_literal(self, literal, context):
        if isinstance(literal, str):
            return literal

        if literal[0] == "number":
            return str(literal[1])

        if literal[0] == "string":
            return stringify(literal[1])

        if literal[0] == "format string":
            return stringify(self._parse_formatted_string(literal[1], context))

        if literal[0] == "actor variable":
            if literal[1].upper() in context.replacements:
                return context.replacements[literal[1].upper()]

            else:
                return literal[1]

        if literal[0] == "call expr":
            return self._parse_action(literal[1], context)

        if literal[0] == "anonymous class":
            return self._parse_anonym_class(literal[1], context)

        if literal[0] == "template derivation":
            return self._parse_template_derivation(literal[1], context)

    def _parse_array(self, arr, context):
        arr = dict(arr)
        arr["value"] = (
            "array",
            [self._parse_expression(e, context) for e in arr["value"][1]],
        )

        return arr

    def _parse_template_derivation(
        self, deriv, context, pending=None, name=None, do_stringify=True
    ):
        template_name, template_parms, deriv_body = deriv

        try:
            template = context.templates[template_name]

        except KeyError:
            raise CompilerError(
                "Unknown template '{}' to derive in {}".format(
                    template_name, context.describe()
                )
            )

        if len(template_parms) != len(template.template_parameters):
            raise CompilerError(
                "Bad number of template parameters for '{}' in {}: expected {}, got {}".format(
                    template_name,
                    context.describe(),
                    len(template.template_parameters),
                    len(template_parms),
                )
            )

        template_parms = [
            self._parse_parameter(a, context, template.template_parameters[i])
            for i, a in enumerate(template_parms)
        ]
        template_labels = {}
        template_body = []
        template_macros = {}
        template_arrays = {}

        for btype, bdata in deriv_body:
            if btype == "array":
                bdata = self._parse_array(bdata, context)
                template_arrays[bdata["name"]] = bdata

            elif btype == "label":
                template_labels[bdata["name"]] = bdata

            elif btype == "macro":
                template_macros[bdata["name"]] = bdata

            template_body.append((btype, bdata))

        new_class = self._derive_class_from_template(
            template,
            template_parms,
            context,
            template_labels,
            template_macros,
            template_arrays,
            template_body,
            pending=pending,
            name=name,
        )

        if do_stringify:
            return stringify(new_class.name)

        else:
            return new_class

    def _parse_action(self, a, context):
        aname = a[0]
        aname = context.replacements.get(aname.upper(), aname)

        args = list(a[1]) if a[1] else []
        args = [
            (context.replacements.get(x.upper(), x) if isinstance(x, str) else x)
            for x in args
        ]
        args = [self._parse_argument(x, context) for x in args]
        args = ", ".join(a for a in args if a)

        return "{}({})".format(aname, args)

    def _parse_state_action_or_body(self, a, context):
        if a[0] == "action":
            return [self._parse_state_action(a[1], context)]

        elif a[0] == "inline body":
            res = []

            for x in a[1]:
                res.extend(self._parse_state_action_or_body(x, context))

            return res

        elif a[0] == "repeated inline body":
            res = []

            cval, xidx, body = a[1]
            count = self._parse_replaceable_number(cval, context)

            if count >= 1:
                for idx in range(count):
                    ctx = context.derive()

                    if xidx:
                        ctx.replacements[xidx] = str(idx)

                    res.extend(self._parse_state_action_or_body(body, ctx))

            return res

    def _parse_state_action(self, a, context):
        args = list(a[1]) if a[1] else []
        args = [self._parse_argument(x, context) for x in args]
        args = [
            (context.replacements.get(x.upper(), x) if isinstance(x, str) else x)
            for x in args
        ]
        args = ", ".join(a for a in args if a)

        if len(args) > 0:
            return "{}({})".format(a[0], args)

        else:
            return a[0]

    def _parse_formatted_string(self, cval, context: ZDCodeParseContext):
        res = []

        for part in cval:
            ptype, pval = part

            if ptype == "str":
                res.append(pval)

            if ptype == "eval":
                res.append(str(self._parse_evaluation(pval, context=context)))

            if ptype == "fmt":
                if pval.upper() in context.replacements:
                    res.append(str(context.replacements[pval.upper()]))

                else:
                    raise CompilerError(
                        "Replacement {} not found while formatting string in {}".format(
                            pval, context.describe()
                        )
                    )

        return "".join(res)

    def _parse_replaceable_number(self, cval, context: ZDCodeParseContext):
        if isinstance(cval, str):
            cval = context.replacements.get(cval.upper(), cval)

        try:
            count = int(cval)

        except ValueError:
            raise CompilerError(
                "Invalid repeat count in {}: expected valid integer, got {}".format(
                    context.describe(), repr(cval)
                )
            )

        else:
            return count

    def _mutate_block(self, mut_func, block, *args):
        return [mut_func(s, *args) for s in block]

    def _maybe_mutate_block(self, mut_func, state, *args):
        state = list(state)

        if state[0] == "ifjump":
            jump, s_yes, s_no = state[1]
            state[1] = (
                jump,
                self._mutate_block(mut_func, s_yes, *args),
                s_no and self._mutate_block(mut_func, s_no, *args),
            )

        elif state[0] == "if":
            cond, s_yes, s_no = state[1]
            state[1] = (
                cond,
                self._mutate_block(mut_func, s_yes, *args),
                s_no and self._mutate_block(mut_func, s_no, *args),
            )

        elif state[0] == "sometimes":
            state[1] = dict(state[1])
            state[1]["body"] = self._mutate_block(mut_func, state[1]["body"], *args)

        elif state[0] == "repeat":
            cval, xidx, body = state[1]
            state[1] = (cval, xidx, self._mutate_block(mut_func, body, *args))

        return state

    def _maybe_mutate_block_or_loop(self, mut_func, state, *args):
        state = list(state)

        if state[0] in ("while", "whilejump"):
            cond, f_body, f_else = state[1]

            f_body = self._mutate_block(mut_func, f_body, *args)
            f_else = f_else and self._mutate_block(mut_func, f_else, *args)

            state[1] = (cond, f_body, f_else)

        elif state[0] == "for":
            itername, iteridx, itermode, f_body, f_else = state[1]

            f_body = self._mutate_block(mut_func, f_body, *args)
            f_else = f_else and self._mutate_block(mut_func, f_else, *args)

            state[1] = (itername, iteridx, itermode, f_body, f_else)

        return self._maybe_mutate_block(mut_func, state, *args)

    def _mutate_macro_state(self, state, inj_context):
        # Mutates each state in an injected macro, making things
        # like macro-scope return statements possible.

        if hasattr(state, "to_decorate"):
            return state.clone()

        if state[0] == "return":
            return ("skip", inj_context)

        return self._maybe_mutate_block_or_loop(
            self._mutate_macro_state, state, inj_context
        )

    def _mutate_iter_state(self, state, break_context, loop_context):
        # Mutates each state in a for loop, making things like
        # break and continue statements possible.

        if state[0] == "continue":
            return ("skip", loop_context)

        if state[0] == "break":
            return ("skip", break_context)

        return self._maybe_mutate_block(
            self._mutate_iter_state, state, break_context, loop_context
        )

    def _parse_state_modifier(self, context: ZDCodeParseContext, modifier_chars):
        res = []

        def recurse(modifier_chars):
            for ctype, cval in modifier_chars:
                if ctype == "replace":
                    try:
                        res.append(context.replacements[cval.upper()])

                    except KeyError:
                        raise CompilerError(
                            "No parameter {} for replacement within modifier, in {}!".format(
                                cval, context.describe()
                            )
                        )

                elif ctype == "recurse":
                    recurse(cval)

                elif ctype == "char":
                    res.append(cval)

        recurse(modifier_chars)

        return "".join(res)

    def _parse_mod_clause(self, context: ZDCodeParseContext, selector, effects):
        return ZDModClause(self, context, selector, effects)

    def _parse_mod_block(self, context: ZDCodeParseContext, mod_block):
        mod_name, mod_body = mod_block
        new_clauses = []

        for selector, effects in mod_body:
            new_clauses.append(self._parse_mod_clause(context, selector, effects))

        context.mods[mod_name.upper()] = new_clauses

    def _parse_state_sprite(self, context: ZDCodeParseContext, sprite):
        sprite_type, sprite_name = sprite

        if sprite_type == "normal":
            name = sprite_name

        elif sprite_type == "parametrized":
            try:
                new_name = context.replacements[sprite_name.upper()]

                if new_name[0] == "'" and new_name[-1] == "'":
                    new_name = new_name[1:-1]

                elif new_name[0] == '"' and new_name[-1] == '"':
                    new_name = new_name[1:-1]

                else:
                    raise CompilerError(
                        "Parametrized sprite '{}' in {} needs to be passed a string; got {}".format(
                            sprite_name, context.describe(), repr(new_name)
                        )
                    )

                name = new_name

            except KeyError:
                raise CompilerError(
                    "No parameter {} for parametrized sprite name, in {}!".format(
                        repr(sprite_name), context.describe()
                    )
                )

        return name

    def _parse_state_expr(self, context: ZDCodeParseContext, s):
        label = ZDDummyLabel(self)
        actor = ZDDummyActor(self, context)

        for state in s:
            self._parse_state(actor, context, label, state)

        return label.states

    def _parse_state(self, actor, context: ZDCodeParseContext, label, s, func=None):
        def add_state(s, target=context):
            added = [s]

            for m in context.get_applied_mods():
                m.apply(context, added)

            target.states.extend(added)
            label.states.extend(added)

            return added

        if hasattr(s, "to_decorate"):
            add_state(s)
            return

        def pop_remote(target=context):
            assert target.remote_children
            target.remote_children.pop()

        def clear_remotes(target=context):
            target.remote_children.clear()

        if s[0] == "frames":
            sprite, frames, duration, modifiers, action = s[1]
            parsed_modifiers = []

            for modifier_chars in modifiers:
                parsed_modifiers.append(
                    self._parse_state_modifier(context, modifier_chars)
                )

            name = self._parse_state_sprite(context, sprite)

            if frames == '"#"':
                frames = ['"#"']

            for f in frames:
                if action is None:
                    add_state(ZDState(name, f, duration, parsed_modifiers))

                else:
                    body = self._parse_state_action_or_body(action, context)

                    for i, a in enumerate(body):
                        add_state(
                            ZDState(
                                name,
                                f,
                                (0 if i + 1 < len(body) else duration),
                                parsed_modifiers,
                                action=a,
                            )
                        )

        elif s[0] == "return":
            raise CompilerError(
                "Return statements are not valid in {}!".format(context.describe())
            )

        elif s[0] == "continue":
            raise CompilerError(
                "Continue statements are not valid in {}!".format(context.describe())
            )

        elif s[0] == "break":
            raise CompilerError(
                "Break statements are not valid in {}!".format(context.describe())
            )

        elif s[0] == "skip":
            # Skips to the end of this context, or the one supplied.
            if s[1]:
                # s[1] can only be an outer context, so don't bother checking
                add_state(ZDSkip(self, s[1], s[1].remote_num_states()))

            else:
                add_state(ZDSkip(self, context, context.remote_num_states()))

        elif s[0] == "call":
            raise CompilerError(
                "Functions and calls have been removed since ZDCode 2.11.0! ({})".format(
                    context.describe()
                )
            )

        elif s[0] == "flow":
            if s[1].upper().rstrip(";") == "LOOP":
                add_state(ZDRawDecorate("goto {}".format(label.name)))

            else:
                sf = s[1].rstrip(";").split(" ")
                sf[0] = sf[0].lower()

                add_state(ZDRawDecorate(" ".join(sf)))

        elif s[0] == "repeat":
            cval, xidx, body = s[1]

            break_ctx = context.derive()
            count = self._parse_replaceable_number(cval, context)

            if count >= 1:
                for idx in range(count):
                    loop_ctx = break_ctx.derive()

                    if xidx:
                        loop_ctx.replacements[xidx.upper()] = str(idx)

                    for a in body:
                        self._parse_state(
                            actor,
                            loop_ctx,
                            label,
                            self._mutate_iter_state(a, break_ctx, loop_ctx),
                            func,
                        )

        elif s[0] == "sometimes":
            s = dict(s[1])

            chance = self._parse_expression(s["chance"], context)
            sms = ZDSometimes(actor, chance, [])

            for a in s["body"]:
                self._parse_state(actor, context, sms, a, func)

            add_state(sms)

        elif s[0] == "apply":
            apply_mod, apply_block = s[1]

            try:
                mod = context.mods[apply_mod.strip().upper()]  # type: List[ZDModClause]

            except KeyError:
                raise CompilerError(
                    "Tried to apply unkown state mod {} in apply statement inside {}!".format(
                        repr(apply_mod), context.describe()
                    )
                )

            apply_ctx = context.derive("apply block")
            apply_ctx.applied_mods.extend(mod)

            apply = ZDBlock(actor)

            for a in apply_block:
                self._parse_state(actor, apply_ctx, apply, a, func)

            add_state(apply)

        elif s[0] == "if":
            ifs = ZDIfStatement(actor, self._parse_expression(s[1][0], context), [])
            if_ctx = context.remote_derive("if body", 3 if s[1][2] else 2)

            for a in s[1][1]:
                self._parse_state(actor, if_ctx, ifs, a, func)

            if s[1][2]:
                elses = ZDBlock(actor)

                for a in s[1][2]:
                    self._parse_state(actor, if_ctx, elses, a, func)

                ifs.set_else(elses)

            add_state(ifs)
            pop_remote()

        elif s[0] == "ifjump":
            jump, s_yes, s_no = s[1]

            @ZDIfJumpStatement.generate(actor)
            def ifs(jump_offset):
                jump_context = context.derive("ifjump check")
                jump_context.replacements["$OFFSET"] = str(jump_offset)

                return self._parse_state_action(jump, jump_context)

            if_ctx = context.remote_derive("ifjump body", 3)

            for a in s_yes:
                self._parse_state(actor, if_ctx, ifs, a, func)

            if s_no:
                elses = ZDBlock(actor)

                for a in s_no:
                    self._parse_state(actor, if_ctx, elses, a, func)

                ifs.set_else(elses)

            add_state(ifs)
            pop_remote()

        elif s[0] == "whilejump":
            break_ctx = context.remote_derive("whilejump", 4)
            jump, s_yes, s_no = s[1]

            @ZDWhileJumpStatement.generate(actor)
            def whs(jump_offset):
                jump_context = break_ctx.derive("whilejump check")
                jump_context.replacements["$OFFSET"] = str(jump_offset)

                return self._parse_state_action(jump, jump_context)

            for a in s_yes:
                loop_ctx = break_ctx.derive("body")
                self._parse_state(
                    actor,
                    loop_ctx,
                    whs,
                    self._mutate_iter_state(a, break_ctx, loop_ctx),
                    func,
                )

            if s_no is not None:
                elses = ZDBlock(actor)

                for a in s_no:
                    self._parse_state(actor, context, elses, a, func)

                whs.set_else(elses)

            add_state(whs)
            clear_remotes(break_ctx)

        elif s[0] == "while":
            break_ctx = context.remote_derive("while", 4 if s[1][2] else 3)
            whs = ZDWhileStatement(
                actor, self._parse_expression(s[1][0], break_ctx), []
            )

            for a in s[1][1]:
                loop_ctx = break_ctx.derive("body")
                self._parse_state(
                    actor,
                    loop_ctx,
                    whs,
                    self._mutate_iter_state(a, break_ctx, loop_ctx),
                    func,
                )

            if s[1][2]:
                elses = ZDBlock(actor)

                for a in s[1][2]:
                    self._parse_state(actor, if_ctx, elses, a, func)

                whs.set_else(elses)

            add_state(whs)
            clear_remotes(break_ctx)

        elif s[0] == "for":
            itername, iteridx, itermode, f_body, f_else = s[1]

            def do_for(iterator):
                break_ctx = context.derive("for")

                for i, item in enumerate(iterator):
                    iter_ctx = break_ctx.derive("{} loop body".format(itermode[0]))
                    iter_ctx.replacements[itername.upper()] = item

                    if iteridx:
                        iter_ctx.replacements[iteridx.upper()] = str(i)

                    break_skip_size = len(f_body) * (
                        len(self.groups[group_name.upper()]) - i
                    )
                    continue_skip_size = len(f_body) * (
                        len(self.groups[group_name.upper()]) - i
                    )

                    for si, a in enumerate(f_body):
                        self._parse_state(
                            actor,
                            iter_ctx,
                            label,
                            self._mutate_iter_state(a, break_ctx, iter_ctx),
                            label,
                        )

            def do_else():
                else_ctx = context.derive("for-else")

                for a in f_else:
                    self._parse_state(actor, else_ctx, label, a, label)

            if itermode[0] == "group":
                group_name = context.resolve(itermode[1], "a parametrized group name")

                if group_name.upper() not in self.groups:
                    raise CompilerError(
                        "No such group {} to 3 in a for loop in {}!".format(
                            repr(group_name), context.describe()
                        )
                    )

                elif self.groups[group_name.upper()]:
                    do_for(iter(self.groups[group_name.upper()]))

                else:
                    do_else()

            elif itermode[0] == "range":
                rang = itermode[1]

                r_from = self._parse_replaceable_number(rang[0], context)
                r_to = rang[0] + self._parse_replaceable_number(rang[1], context)

                if r_to > r_from:
                    do_for(list(range(r_from, r_to, 1)))

                else:
                    do_else()

            else:
                raise CompilerError(
                    "Unknown internal for loop iteration mode '{}' in {}! Please report this issue to the author.".format(
                        itermode[0], context.describe()
                    )
                )

        elif s[0] == "inject":
            r_from, r_name, r_args = s[1]
            r_name = context.resolve(r_name, "a parametrized macro injection")

            if r_from:
                r_from = unstringify(
                    context.resolve(r_from, "a parametrized extern macro classname")
                )

                if r_from.upper() in self.actor_names:
                    act = self.actor_names[r_from.upper()]

                else:
                    raise CompilerError(
                        "Unknown extern macro classname {} in {}!".format(
                            repr(r_from), context.describe()
                        )
                    )

                macros = dict(act.context.macros)

            else:
                macros = dict(context.macros)

            if r_name.upper() in macros:
                if r_from:
                    new_context = context.derive(
                        "macro '{}' from {}".format(r_name, act.name)
                    )
                    new_context.update(act.context)
                    r_qualname = "{}.{}".format(r_from, r_name)

                else:
                    new_context = context.derive("macro '{}'".format(r_name))
                    r_qualname = r_name

                (m_args, m_body) = macros[r_name.upper()]

                if len(m_args) != len(r_args):
                    raise CompilerError(
                        "Bad number of arguments while trying to inject macro {}; expected {}, got {}! (in {})".format(
                            r_qualname, len(m_args), len(r_args), context.describe()
                        )
                    )

                for rn, an in zip(r_args, m_args):
                    new_context.replacements[an.upper()] = self._parse_argument(
                        rn, context, an
                    )

                for a in m_body:
                    self._parse_state(
                        actor,
                        new_context,
                        label,
                        self._mutate_macro_state(a, new_context),
                        label,
                    )

            else:
                if r_from:
                    raise CompilerError(
                        "Unknown macro {}.{} in {}!".format(
                            r_from, r_name, context.describe()
                        )
                    )

                else:
                    raise CompilerError(
                        "Unknown macro {} in {}!".format(r_name, context.describe())
                    )

    def _parse_inherit(self, inh, context):
        if inh is None:
            return None

        ptype, pval = inh

        if ptype == "classname":
            return context.replacements.get(pval.upper(), pval)

        elif ptype == "template derivation":
            with context.desc_block("template derivation inheritance"):
                return self._parse_template_derivation(
                    pval, context, do_stringify=False
                ).name

    def _parse_anonym_class(self, anonym_class, context):
        a = dict(anonym_class)
        new_context = context.derive("anonymous class")

        classname = "_AnonymClass_{}_{}".format(self.id, len(self.anonymous_classes))

        if a["group"]:
            g = unstringify(a["group"])

            if g.upper() in self.groups:
                self.groups[g.upper()].append(stringify(classname))

            else:
                raise CompilerError(
                    "Group '{}' not found while compiling anonymous class in {}!".format(
                        g, context.describe()
                    )
                )

        anonym_actor = ZDActor(
            self,
            classname,
            inherit=self._parse_inherit(a["inheritance"], context),
            context=new_context,
        )

        self._parse_class_body(anonym_actor, anonym_actor.get_context(), a["body"])

        context.add_actor(anonym_actor)

        self.anonymous_classes.append(anonym_actor)
        self.inventories.append(anonym_actor)

        return stringify(anonym_actor.name)

    def _derive_class_from_template(
        self,
        template,
        param_values,
        context,
        labels=(),
        macros=(),
        arrays=(),
        body=(),
        name=None,
        pending=None,
    ):
        labels = dict(labels)
        macros = dict(macros)
        arrays = dict(arrays)
        body = list(body)

        name = name or template.generated_class_name(param_values, make_id(40))

        needs_init, actor = template.generate_init_class(
            self,
            context,
            param_values,
            {l.upper() for l in labels.keys()},
            {m["name"].upper(): m["args"] for m in macros.values()},
            {a["name"].upper(): len(a["value"]) for a in arrays.values()},
            name=name,
        )
        new_context = actor.get_context()

        if needs_init:

            def pending_oper_gen():
                act = actor
                new_ctx = new_context
                temp = template

                @pending_task(0)
                def pending_oper():
                    for btype, bdata in body:
                        if btype == "array":
                            try:
                                absarray = template.abstract_array_names[
                                    bdata["name"].upper()
                                ]

                            except KeyError:
                                raise CompilerError(
                                    "Tried to define an array {} in {} that is not abstractly declared in the template {}!".format(
                                        repr(bdata["name"]),
                                        new_ctx.describe(),
                                        repr(template.name),
                                    )
                                )

                            act.uservars.append(
                                {
                                    "name": bdata["name"],
                                    "size": len(bdata["value"][1]),
                                    "value": ("arr", bdata["value"][1]),
                                    "type": absarray["type"],
                                }
                            )

                    self._parse_class_body(act, new_ctx, temp.parse_data + body)

                return pending_oper

            if pending:
                pending.put_nowait(pending_oper_gen())

            else:
                pending_oper_gen().func()

            self.actors.append(actor)
            self.actor_names[actor.name.upper()] = actor

        return actor

    def _parse_class_body(self, actor, context, body):
        assert context is actor.context

        for btype, bdata in body:
            if btype == "mod":
                self._parse_mod_block(context, bdata)

        for btype, bdata in body:
            if btype == "macro":
                context.macros[bdata["name"].upper()] = (bdata["args"], bdata["body"])

        for btype, bdata in body:
            if btype == "property":
                ZDProperty(
                    actor,
                    bdata["name"],
                    ", ".join(
                        self._parse_parameter(x, context) for x in bdata["value"]
                    ),
                )

            elif btype == "flag":
                actor.flags.add(bdata)

            elif btype == "flag combo":
                actor.raw.append(bdata)

            elif btype == "user var":
                bdata = {
                    **bdata,
                    "value": self._parse_expression(bdata["value"], context),
                }
                actor.uservars.append(bdata)

            elif btype == "unflag":
                actor.antiflags.add(bdata)

            elif btype == "label":
                label = ZDLabel(actor, bdata["name"])

                with context.desc_block("label '{}'".format(label.name)):
                    for s in bdata["body"]:
                        self._parse_state(actor, context, label, s, None)

            elif btype == "function":
                raise CompilerError(
                    "Functions have been removed since ZDCode 2.11.0! ({})".format(
                        context.describe()
                    )
                )

            elif btype == "apply":
                mtype, mval = bdata

                if mtype == "name":
                    try:
                        mod = context.mods[
                            mval.strip().upper()
                        ]  # type: List[ZDModClause]

                    except KeyError:
                        raise CompilerError(
                            "Tried to apply unkown state mod {} in global apply statement inside {}!".format(
                                repr(mval.strip()), context.describe()
                            )
                        )

                elif mtype == "body":
                    mod = []

                    for selector, effects in mval:
                        mod.append(self._parse_mod_clause(context, selector, effects))

                context.applied_mods.extend(mod)

    def _parse(self, actors):
        parsed_actors = []

        context = ZDCodeParseContext(actors=[parsed_actors])

        actors = [(context, a) for a in actors]

        # unpack static for loops
        def resolve_for(forbody):
            itername, iteridx, itermode, f_body, f_else = forbody

            def do_for(iterator):
                break_ctx = context.derive("for")

                for i, item in enumerate(iterator):
                    iter_ctx = break_ctx.derive("for..{} loop body".format(itermode[0]))
                    iter_ctx.replacements[itername.upper()] = item

                    if iteridx:
                        iter_ctx.replacements[iteridx.upper()] = str(i)

                    for si, a in enumerate(f_body):
                        yield iter_ctx, a

            def do_else():
                else_ctx = context.derive("for-else")

                for a in f_else:
                    yield else_ctx, a

            if itermode[0] == "group":
                group_name = context.resolve(itermode[1], "a parametrized group name")

                if group_name.upper() not in self.groups:
                    raise CompilerError(
                        "No such group {} to 3 in a for loop in {}!".format(
                            repr(group_name), context.describe()
                        )
                    )

                elif self.groups[group_name.upper()]:
                    yield from do_for(iter(self.groups[group_name.upper()]))

                else:
                    yield from do_else()

            elif itermode[0] == "range":
                rang = itermode[1]

                r_from = self._parse_replaceable_number(rang[0], context)
                r_to = rang[1][0] + self._parse_replaceable_number(rang[1][1], context)

                if r_to > r_from:
                    r = list(range(r_from, r_to, 1))
                    yield from do_for(r)

                else:
                    yield from do_else()

        def unpack(vals, reslist):
            # always parse groups first,
            # to ensure they can be seen by
            # static for loops
            unpacked = True

            for ctx, (class_type, a) in vals:
                if class_type == "group":
                    a = dict(a)

                    gname = a["name"].upper()

                    if gname in self.groups:
                        self.groups[gname].extend(list(a["items"]))

                    else:
                        self.groups[gname] = list(a["items"])

            for ctx, (class_type, a) in vals:
                if class_type == "for":
                    unpacked = False
                    reslist.extend(resolve_for(a))

                elif class_type != "group":
                    reslist.append((ctx, (class_type, a)))

            return unpacked

        while True:
            reslist = []

            if unpack(actors, reslist):
                break

            else:
                actors = reslist
                reslist = []

        # read the unpacked list
        for ctx, (class_type, a) in actors:
            if class_type == "macro":
                a = dict(a)

                context.macros[a["name"]] = (a["args"], a["body"])

        for actx, (class_type, a) in actors:
            if class_type == "class template":
                a = dict(a)

                abstract_labels = set()
                abstract_macros = {}
                abstract_arrays = {}
                template_body = []

                if a["group"]:
                    g = unstringify(a["group"])

                    if g.upper() not in self.groups:
                        raise CompilerError(
                            "Group '{}' not found while compiling template class {}!".format(
                                g, a["classname"]
                            )
                        )

                    g = g.upper()

                else:
                    g = None

                for btype, bdata in a["body"]:
                    if btype == "abstract label":
                        abstract_labels.add(bdata.upper())

                    elif btype == "abstract macro":
                        abstract_macros[bdata["name"].upper()] = bdata["args"]

                    elif btype == "abstract array":
                        abstract_arrays[bdata["name"].upper()] = bdata

                    else:
                        template_body.append((btype, bdata))

                template = ZDClassTemplate(
                    a["parameters"],
                    template_body,
                    abstract_labels,
                    abstract_macros,
                    abstract_arrays,
                    g,
                    self,
                    a["classname"],
                    self._parse_inherit(a["inheritance"], actx),
                    a["replacement"],
                    a["class number"],
                )
                actx.templates[a["classname"]] = template

        pending = queue.PriorityQueue()

        for actx, (class_type, a) in actors:
            if class_type == "class":
                a = dict(a)

                with actx.desc_block("class '{}'".format(a["classname"])):
                    actor = ZDActor(
                        self,
                        a["classname"],
                        self._parse_inherit(a["inheritance"], actx),
                        a["replacement"],
                        a["class number"],
                        context=actx,
                    )
                    ctx = actor.get_context()

                    if a["group"]:
                        g = a["group"]

                        if g.upper() in self.groups:
                            self.groups[g.upper()].append(stringify(a["classname"]))

                        else:
                            raise CompilerError(
                                "Group '{}' not found while compiling class '{}'!".format(
                                    g, a["classname"]
                                )
                            )

                    def pending_oper_gen():
                        actor_ = actor
                        ctx_ = ctx
                        body_ = a["body"]

                        @pending_task(2)
                        def pending_oper():
                            self._parse_class_body(actor_, ctx_, body_)

                        return pending_oper

                    pending.put_nowait(pending_oper_gen())

                    self.actors.append(actor)
                    self.actor_names[actor.name.upper()] = actor
                    parsed_actors.append(actor)

            elif class_type == "static template derivation":
                a = dict(a)

                new_name = a["classname"]
                gname = a["group"]

                with actx.desc_block(
                    "static template derivation '{}'".format(new_name)
                ):
                    parsed_actors.append(
                        self._parse_template_derivation(
                            a["source"][1],
                            actx,
                            pending=pending,
                            name=new_name,
                            do_stringify=False,
                        )
                    )

                    if gname:

                        @pending_task(1)
                        def pending_oper():
                            g = unstringify(gname)

                            if g.upper() not in self.groups:
                                raise CompilerError(
                                    "No such group {} to add the derivation {} to!".format(
                                        repr(g), new_name
                                    )
                                )

                            self.groups[g.upper()].append(new_name)

                        pending.put_nowait(pending_oper)

        while not pending.empty():
            pending.get_nowait().func()

        for a in parsed_actors:
            a.prepare_spawn_label()

    def __init__(self):
        self.includes = {}
        self.inventories = []
        self.anonymous_classes = []
        self.actors = []
        self.actor_names = {}
        self.groups = {}
        self.id = make_id(35)
        self.num_anonym_macros = 0

    def to_decorate(self):
        res = TextNode(indent=0)

        if self.inventories:
            for i in self.inventories:
                res.add_line(i.to_decorate())

        for a in self.actors:
            res.add_line(a.to_decorate())

        return res

    def decorate(self):
        return decorate(self)



import heapq
import sys
import typing
from configparser import ConfigParser, ExtendedInterpolation
from enum import Enum

import attr

try:
    from zdcode.bundle import Bundle, BundleOutput

except ImportError:
    from bundle import Bundle, BundleOutput


class ZakeException(Exception):
    pass


class ConfigError(ZakeException):
    pass


class ZakeTarget:
    def __init__(self, name: str):
        self.name = name
        self.definitions: dict[str, str] = {}
        self.inputs: list[tuple[str, str]] = []
        self.outputs: dict[str, BundleOutput] = {}

    def add_input(self, inp):
        self.inputs.append(inp)

    def add_definition(self, key, value):
        self.definitions[key] = value

    def bundle(self, **kwargs):
        bundle = Bundle(*self.inputs, outputs=self.outputs)

        return bundle.bundle(
            preproc_defs={k.upper(): v for k, v in self.definitions.items()}, **kwargs
        )


class Zake:
    def __init__(self):
        self.targets: dict[str, ZakeTarget] = {}

    def add_target(self, name: str) -> ZakeTarget:
        return self.targets.setdefault(name, ZakeTarget(name))

    def read(self, fn, **kwargs):
        config = ConfigParser(
            interpolation=ExtendedInterpolation(), **kwargs, strict=False
        )
        config.read(fn)

        c_general = config["General"]

        # general values
        name = c_general["name"].strip()
        version = c_general["version"].strip()
        targets = c_general["targets"].strip().split()

        if "partitions" in c_general:
            bundles = [x.lower() for x in c_general["partitions"].strip().split()]

        else:
            bundles = ["asset", "code"]

        # set values (for interpolation)
        interp = {"name": name, "version": version}

        # default values
        targs = {}

        def get_bundle_cfg(s_bundle_cfg, name):
            return config.get(
                s_bundle_cfg,
                name,
                vars=interp,
                fallback=config.get("Partition", name, vars=interp, fallback=None),
            )

        # targets
        for tname in targets:
            targ = self.add_target(tname)
            targs[tname] = targ

            interp["target"] = tname  # for interpolation
            interp["defaults"] = ""

            s_pats = "Paths." + tname
            s_defs = "Definitions." + tname
            s_bundle_cfg = "Partition." + tname

            # define bundle outputs
            for bundle in bundles:
                targ.outputs[bundle] = BundleOutput(
                    name=bundle,
                    output=None,
                )

            if "partitions" not in c_general:
                targ.outputs["asset"].priority = 1.0
                targ.outputs["code"].priority = 100.0
                targ.outputs["code"].add_matcher("DECORATE")
                targ.outputs["code"].add_matcher("DECORATE*")

            # section values
            pats = dict(config.items("Paths", vars=interp)) if "Paths" in config else {}

            defs = (
                dict(config.items("Definitions", vars=interp))
                if "Definitions" in config
                else {}
            )

            if s_pats in config:
                new_pat_keys = dict(config.items(s_pats, vars=interp)).keys()

                for key in new_pat_keys:
                    interp["defaults"] = ""

                    if key in pats:
                        interp["defaults"] = pats[k]

                    pats[key] = config.get(s_pats, key, vars=interp)

            if s_defs in config:
                defs.update(config.items(s_defs, vars=interp))

            # fetch inputs and outputs
            if "inputs" not in pats:
                raise ConfigError(
                    "Required Zake field 'inputs' missing from section Paths while reading target {}.".format(
                        tname
                    )
                )

            for inp in pats["inputs"].strip().split():
                targ.add_input((inp, ""))

            if "injects" in pats:
                for injs in pats["injects"].strip().split():
                    inp, out = injs.split()
                    targ.add_input((inp, out))

            for bundle in bundles:
                out_name = "output." + bundle
                matchers_name = "matchers." + bundle
                excluders_name = "excluders." + bundle
                priority_name = "priority." + bundle

                priority = config.getfloat(
                    s_bundle_cfg,
                    priority_name,
                    vars=interp,
                    fallback=config.getfloat(
                        "Partition", priority_name, vars=interp, fallback=None
                    ),
                )

                if priority is not None:
                    targ.outputs[bundle].priority = priority

                if out_name in pats:
                    targ.outputs[bundle].output = pats[out_name].strip()

                else:
                    raise ConfigError(
                        "Required Zake field '{}' missing from section Paths while reading target {}.".format(
                            out_name, tname
                        )
                    )

                matchers = get_bundle_cfg(s_bundle_cfg, matchers_name)
                excluders = get_bundle_cfg(s_bundle_cfg, excluders_name)

                if matchers:
                    for matcher in matchers:
                        targ.outputs[bundle].add_matcher(matcher)

                if excluders:
                    for excluder in excluders:
                        targ.outputs[bundle].add_excluder(excluder)

            # preprocessor definitions
            for k, v in defs.items():
                targ.add_definition(k.strip(), v.strip())

        # return parsed targets
        return targs

    def execute(self, **kwargs):
        print(
            "Starting ZDCode bundling barrage with {} targets.".format(
                len(self.targets)
            )
        )

        for name, target in self.targets.items():
            print("\n-- Bundling target: " + name)
            yield name, target.bundle(**kwargs)


# main function for Zake
def main(print_status_code=True):
    filename = "Zake.ini"

    if len(sys.argv) > 1:
        filename = sys.argv.pop(1)

    zake = Zake()
    zake.read(filename)

    acc_status = None

    total = 0
    success = 0
    error = 0

    for tname, (status, message) in zake.execute():
        if acc_status is None:
            acc_status = status

        else:
            acc_status = max(status, acc_status)

        if status == 0:
            success += 1

        else:
            error += 1

        total += 1

        print(" - {}: {}".format(tname, message))

    print(
        "{tot} targets processed ({successes}).{status}".format(
            tot=total,
            successes=(
                "{} successful, {} failed".format(success, error)
                if error
                else ("all successful" if success else "nothing done")
            ),
            status=(
                " Final status {}.".format(acc_status) if print_status_code else ""
            ),
        )
    )

    exit(acc_status)


if __name__ == "__main__":
    main()



try:
    from . import ZDCode

except ImportError:
    from zdcode import ZDCode

import fnmatch
import functools
import heapq
import pathlib
import shutil
import subprocess
import tempfile
import typing
import zipfile
from collections import deque

import attr


@functools.total_ordering
@attr.s(order=False)
class BundleOutput:
    name: str = attr.ib()
    output: str = attr.ib()
    priority: float = attr.ib(default=1.0)
    matchers: list[str] = attr.ib(factory=list)
    excluders: list[str] = attr.ib(factory=list)

    def add_matcher(self, matcher: str):
        self.matchers.append(matcher)

    def add_excluder(self, excluder: str):
        self.excluders.append(excluder)

    def matches(self, rel_path: str) -> bool:
        for excluder in self.excluders:
            if fnmatch.fnmatch(rel_path, excluder):
                return False

        if not self.matchers:
            return True

        for matcher in self.matchers:
            if fnmatch.fnmatch(rel_path, matcher):
                return True

        return False

    def __gt__(self, other: "BundleOutput") -> bool:
        return self.priority < other.priority

    def __eq__(self, other: "BundleOutput") -> bool:
        return self.priority == other.priority


class BundleDependencyError(Exception):
    pass


@attr.s
class BundleInputWalker:
    code: ZDCode = attr.ib()
    bundle: "Bundle" = attr.ib()
    deps: list[str] = attr.ib(factory=list)
    bundled: set[str] = attr.ib(factory=set)
    build_tasks: list[typing.Callable[[], bool]] = attr.ib(factory=list)
    error_handler: typing.Optional[
        typing.Callable[[ZDCode.ZDCodeError], None]
    ] = attr.ib(default=None)
    preproc_defs: dict[str, str] = attr.ib(factory=dict)
    collected: typing.Deque[tuple[str, str, bytes]] = attr.ib(factory=list)

    @classmethod
    def new(self, bundle, error_handler=None, preproc_defs=None) -> "BundleInputWalker":
        return BundleInputWalker(
            bundled=set(),
            deps=[],
            code=ZDCode(),
            build_tasks=[],
            error_handler=error_handler,
            preproc_defs=preproc_defs or {},
            bundle=bundle,
            collected=deque(),
        )

    def add_dep(self, url: pathlib.Path, target: pathlib.PurePath) -> ():
        self.deps.append((url, target))

    def scan_deps(self):
        while self.deps:
            mod, target = self.deps.pop()
            self.scan_dep(mod, target)

    def build(self) -> typing.Optional[tuple[int, str]]:
        while self.build_tasks:
            task = self.build_tasks.pop()

            if not task():
                return (
                    1,
                    "Errors were found during the compilation of the ZDCode lumps!",
                )

        self.collect(pathlib.PurePath("DECORATE"), self.code.decorate().encode("utf-8"))

        return None

    def store_collected(self, out_zip: zipfile.ZipFile, target: str, data: bytes):
        # check if file already exists in target;
        # if so, add extension

        # (ignore folders)

        out_path = target
        num = 0

        names_list = out_zip.namelist()

        while out_path in names_list:
            num += 1
            out_path = target + "." + str(num)

        out_zip.writestr(out_path, data)

    def assemble(self):
        zips: dict[str, zipfile.ZipFile] = {}
        zipfiles = set()

        for oname, out in self.bundle.outputs.items():
            oname = oname.lower()

            zipf = zipfile.ZipFile(out.output, mode="w")
            zips[oname] = zipf
            zipfiles.add(zipf)

        while self.collected:
            target, oname, data = self.collected.popleft()
            out_zip = zips[oname.lower()]
            self.store_collected(out_zip, target, data)

    def scan_dep(self, url: pathlib.Path, target: pathlib.PurePath):
        url_str = str(url)

        if url_str in self.bundled:
            return

        self.bundled.add(url_str)
        self.scan_dep_url(url, target, url)

    def scan_dep_url(
        self, url: pathlib.Path, target: pathlib.PurePath, relative: pathlib.PurePath
    ):
        if url.is_dir():
            self.scan_dep_dir(url, target, relative)

        elif url.is_file():
            self.scan_dep_file(url, target, relative)

    def scan_dep_zip(
        self, url: pathlib.Path, target: pathlib.PurePath, relative: pathlib.PurePath
    ):
        with tempfile.TemporaryDirectory() as extract_out:
            extractdest = pathlib.Path(extract_out)

            with zipfile.ZipFile(url) as zipped:
                zipped.extractall(extractdest)

            self.scan_dep_dir(extractdest, target, relative)

    def scan_dep_file(
        self, url: pathlib.Path, target: pathlib.PurePath, relative: pathlib.PurePath
    ):
        if url.stem.split(".")[0].upper() == "ZDCODE" or url.suffix.upper() == ".ZC2":
            self.build_tasks.append(self._compile_task(url))

        elif url.suffix.upper() in (".PK3", ".PKZ", ".ZIP"):
            self.scan_dep_zip(url, target, relative)

        else:
            opath = target / url.name
            self.collect(opath, url.read_bytes())

    def collect(self, out_path: pathlib.PurePath, data: bytes):
        output = self.bundle.find_output(out_path.name)

        if output:
            self.collected.append((str(out_path), output.name.lower(), data))

    def _compile_task(self, zdc):
        def compile_mod_zdcode():
            with zdc.open() as zdc_fp:
                return self.code.add(
                    zdc_fp.read(),
                    zdc.name,
                    zdc.parent,
                    self.error_handler,
                    preproc_defs=self.preproc_defs,
                )

        return compile_mod_zdcode

    def scan_dep_dir(
        self, url: pathlib.Path, target: pathlib.PurePath, relative: pathlib.PurePath
    ):
        for filepath in url.rglob("*"):
            if filepath.is_file() and filepath.name != "DEPINDEX":
                self.scan_dep_file(
                    filepath, target / filepath.parent.relative_to(url), relative
                )

        indx_path = url / "DEPINDEX"

        if indx_path.is_file():
            lines = indx_path.read().splitlines()

            for l in lines:
                l = l.strip()

                if l:
                    dep_path = url.parent / l

                    if not dep_path.exists():
                        dep_path = pathlib.Path.cwd() / l

                    if not dep_path.exists():
                        raise BundleDependencyError(
                            "The file {} depends on {}, which does not exist, neither as a sibling of the dependent's dir, nor under the working one!"
                        )

                    self.deps.append(dep_path)


class Bundle:
    def __init__(self, *mods: list[tuple[str, str]], outputs=None, error_handler=None):
        self.mods = list(mods)
        self.error_handler = error_handler

        self.outputs: dict[str, BundleOutput] = outputs or {}

        self.output_heap: list[BundleOutput] = list(self.outputs.values())
        self.output_heap.sort()

    def add_output(self, name: str, output: str, priority: float = 1.0) -> BundleOutput:
        res = self.outputs.setdefault(
            name, BundleOutput(name=name, output=output, priority=priority)
        )

        if res not in self.output_heap:
            heapq.heappush(self.output_heap, res)
            self.output_heap.sort()

        return res

    def find_output(self, rel_path: str) -> typing.Optional[BundleOutput]:
        for bundle in self.output_heap:
            if bundle.matches(rel_path):
                return bundle

        return None

    def bundle(
        self,
        error_handler=None,
        preproc_defs=(),
    ):
        walker = BundleInputWalker.new(
            error_handler=error_handler or self.error_handler,
            preproc_defs=dict(preproc_defs),
            bundle=self,
        )

        for mod, modtarg in self.mods:
            walker.add_dep(pathlib.Path(mod), pathlib.Path(modtarg))

        # scan input paths
        print("Scanning inputs...")
        walker.scan_deps()

        # build any zdcode found into decorate
        print("Building ZDCode...")
        err = walker.build()

        if err:
            return err

        # count files bundled
        print("Collected {} files.".format(len(walker.collected)))

        # assemble outputs
        print("Assembling...")
        walker.assemble()

        return (0, "Success!")



# pylint: disable=unreachable
import glob
import os
import re
import sys
import traceback

import parsy

try:
    import simplejson as json

except ImportError:
    import json

from parsy import fail, generate, regex, seq, string, success, whitespace

s = string
fa = fail
whitespace = whitespace.desc('whitespace')
wo = whitespace.optional()

def istring(st):
    return s(st, transform=lambda s: s.upper())

ist = istring

string_body_1 = regex(r'[^"\\]+')
string_body_2 = regex(r'[^\'\\]+')
string_esc = string('\\') >> (
    string('\\')
    | string('/')
    | string('"')
    | string('t').result('\t')
    | regex(r'u[0-9a-fA-F]{4}').map(lambda s: chr(int(s[1:], 16)))
    | regex(r'x[0-9a-fA-F]{2}').map(lambda s: chr(int(s[1:], 16)))
)
string_delim_1 = s('"')
string_delim_2 = s("'")
string_body_plex_1 = (string_body_1 | string_esc).many().concat()
string_body_plex_2 = (string_body_2 | string_esc).many().concat()
string_literal = (
    (string_delim_1 >> string_body_plex_1 << string_delim_1)
    | (string_delim_2 >> string_body_plex_2 << string_delim_2)
)
lwhitespace = whitespace | s('\n') | s('\r')

ifneq = re.compile(r'^\#IF(N|NOT)EQ(UALS?)?$')
ifundef = re.compile(r'^\#IF(U?N|NOT)DEF(INED)?$')
defmacro = re.compile(r'^\#DEF(INE)?M(AC(RO)?)?$')

error_line = re.compile(r'(.+?) at (\d+)\:(\d+)$')

class PreprocessingError(BaseException):
    def __init__(self, problem, line, fname = None, line_content = None):
        self.problem = problem
        self.error_line = line
        self.filename = fname
        self.line_content = line_content

    def __str__(self):
        return "{} (line {} {}){}".format(self.problem, self.error_line, 'of input code' if self.filename is None else 'in "{}"'.format(self.filename), '' if self.line_content is None else ':\n> ' + self.line_content)

class ZDParseError(BaseException):
    def __init__(self, parsy_error: parsy.ParseError, postline=None):
        self.parsy_error = parsy_error
        self.postline = postline

    def __str__(self):
        filename = self.postline[0]
        error_line = self.postline[1]
        line_source = self.postline[2]

        return "expected one of {} at line {}{} {}:\n> {}\n{}^".format(', '.join(repr(l) for l in self.parsy_error.expected), error_line, ':' + str(self.parsy_error.index - self.postline[4]), 'of input code' if filename is None else 'in "{}"'.format(filename), line_source, ' ' * max(0, self.parsy_error.index - self.postline[4] + 2))

@generate
def state_modifier_name():
    return (s('{') >> regex(r'[a-zA-Z_][a-zA-Z_0-9]*').desc('modifier parameter name').tag('replace') << s('}') | (ist('(').tag('char') + state_modifier_name.tag('recurse') + ist(')').tag('char')) | regex(r'[a-zA-Z0-9\'\",\w"]').tag('char')).many().desc('state keyword')
    yield

@generate
def modifier():
    # State modifier. Not to be confused with the 'mod' modifier block.
    return string('[').then(state_modifier_name).skip(s(']'))
    yield

def _apply_replacements(expr, replacements):
    res = []

    for item in expr:
        if item in replacements:
           res.append(replacements[item])

        else:
            res.append(item)

    return ' '.join(res)

def apply_replacements(expr, replacements):
    return _apply_replacements(expression.parse(expr), replacements)

@generate
def number_lit():
    return (
            regex(r'[\+\-]?\d+(.\d+)?(e[\+\-]?\d+)?').tag('number')
        |   regex(r'0x[0-9A-Fa-f]+').map(int).tag('number')
        |   regex(r'0b[01]+').map(int).tag('number')
        )
    yield

@generate
def group_name():
    return regex(r'[a-zA-Z0-9_]+').desc('group name')
    yield
    
@generate
def p_group_name():
    return regex(r'\@*[a-zA-Z0-9_]+').desc('group name')
    yield

@generate
def p_range_vals():
    return seq(
        (replaceable_number | success(0)) << wo,
        ist('..') >>\
        wo >> (s('=') >> wo >> replaceable_number).tag(1) | replaceable_number.tag(0)
    )
    yield

@generate
def variable_name():
    return regex(r'[a-zA-Z_\$][a-zA-Z0-9_\$\[\]]*')
    yield

@generate
def format_string_literal():
    return (
        ist('f') >> wo >> s('{') >> wo >> (
            (
                string_literal.tag('str') |
                numeric_eval.tag('eval') |
                variable_name.tag('fmt')
            ).sep_by(wo)
        ) << wo << s('}')
    )
    yield

eval_literal = regex('[\-\+]?\d+()').map(int) | regex('[\-\+]?\d*(\.\d*)?([Ee]\d+)?').map(float)

@generate
def eval_body():
    return wo >> (
        (ist('(') >> eval_body << ist(')')) |
        eval_literal.tag('literal') | eval_operation.tag('operation')
    ) << wo
    yield

# operator precedence based on C
def opmap(operands, func):
    if operands == 1:
        return lambda _, a: (func, (a,))

    else:
        return lambda a, _, b: (func, (a, b))

operators = [
    # Unary sign precedence
    seq(eval_body, ist('+') << wo, eval_body).map(opmap(1, lambda a: +a)),
    seq(eval_body, ist('-') << wo, eval_body).map(opmap(1, lambda a: -a)),

    # Multiplicative precedence
    seq(eval_body, wo >> ist('%') << wo, eval_body).map(opmap(2, lambda a, b: a % b)),
    seq(eval_body, wo >> ist('*') << wo, eval_body).map(opmap(2, lambda a, b: a * b)),
    seq(eval_body, wo >> ist('/') << wo, eval_body).map(opmap(2, lambda a, b: a / b)),

    # Additive precedence
    seq(eval_body, wo >> ist('+') << wo, eval_body).map(opmap(2, lambda a, b: a + b)),
    seq(eval_body, wo >> ist('-') << wo, eval_body).map(opmap(2, lambda a, b: a - b)),

    # Bit shift precedence
    seq(eval_body, wo >> ist('>>') << wo, eval_body).map(opmap(2, lambda a, b: a >> b)),
    seq(eval_body, wo >> ist('<<') << wo, eval_body).map(opmap(2, lambda a, b: a << b)),

    # Bitwise operation precedence
    seq(eval_body, wo >> ist('&') << wo, eval_body).map(opmap(2, lambda a, b: a & b)),
    seq(eval_body, wo >> ist('^') << wo, eval_body).map(opmap(2, lambda a, b: a ^ b)),
    seq(eval_body, wo >> ist('|') << wo, eval_body).map(opmap(2, lambda a, b: a | b)),

    # Logical (1 or 0) operation precedence
    seq(eval_body, wo >> ist('&&') << wo, eval_body).map(opmap(2, lambda a, b: 1 if (a and b)               else 0)),
    seq(eval_body, wo >> ist('||') << wo, eval_body).map(opmap(2, lambda a, b: 1 if (a or  b)               else 0)),
    seq(eval_body, wo >> ist('^^') << wo, eval_body).map(opmap(2, lambda a, b: 1 if ((a == 0) != (b == 0))  else 0)),

    # Ternary operation precedence
    seq(eval_body, wo >> ist('?') << wo, eval_body, wo >> ist(':') << wo, eval_body).map(
        lambda cond, _, yes, _2, no: (
            (lambda cond, yes, no: yes if cond else no),
            (cond, yes, no)
        )
    ),

    # Comma operation precedence
    seq(eval_body, ist(','), eval_body).map(opmap(2, lambda a, b: b)),
]

@generate
def eval_operation():
    res = None

    for o in operators:
        if res is None:
            res = o

        else:
            res = res | o

    return res
    yield

@generate
def numeric_eval():
    return (
        ist('e') >> wo >> s('{') >> eval_body << s('}')
    )
    yield

@generate
def literal():
    return (
        call_literal.tag('call expr').desc('call') |
        format_string_literal.tag('format string') |
        numeric_eval.tag('eval').desc('numerical evaluation') |
        string_literal.tag('string').desc('string') |
        variable_name.tag('actor variable').desc('actor variable') |
        number_lit.desc('number')
    )
    yield

@generate
def array_literal():
    return (wo >> s('{') >> wo >> expression.sep_by(s(',') << wo) << s('}')).tag('array')
    yield

@generate
def paren_expr():
    return s('(') >> expression << s(')')
    yield

@generate
def expression():
    return (
        (
            wo >>
            (
                (
                    paren_expr.tag('paren expr').desc('parenthetic expression') |
                    literal.tag('literal') |
                    regex(r'[\+\-\|\>\<\~\&\!\=\*\/\%\[\]]+').desc('operator').tag('oper')
                ).sep_by(wo).tag('expr')
            )
            << wo
        ) | (wo >> paren_expr.tag('paren expr') << wo)
    )
    yield


@generate
def arg_expression():
    return parameter.tag('position arg').desc('positional argument value')
    yield

@generate
def argument_list():
    return (literal).sep_by(regex(r',\s*'))
    yield

@generate
def macro_argument_list():
    return wo >> regex(r'[a-zA-Z_][a-zA-Z_0-9]*').desc('macro argument name').sep_by(regex(r',\s*')) << wo
    yield

@generate
def template_parameter_list():
    return wo >> regex(r'[a-zA-Z_][a-zA-Z_0-9]*').desc('template parameter name').sep_by(regex(r',\s*')) << wo
    yield

@generate
def expr_argument_list():
    return arg_expression.sep_by(s(',') >> wo)
    yield

@generate
def parameter():
    return anonymous_class | anonymous_macro | templated_class_derivation | expression.tag('expression')
    yield

def specialized_parameter_list(ptype):
    @generate
    def plist():
        return ptype.sep_by(s(',') >> wo, min=1)
        yield

    return plist

parameter_list = specialized_parameter_list(parameter)

@generate
def actor_function_call():
    return ist('call ').desc("'call' statement").then(regex(r'[a-zA-Z_][a-zA-Z_0-9]*').desc('called function name'))
    yield

@generate
def state_call():
    return seq(
        regex(r'[a-zA-Z_][a-zA-Z_0-9]*').desc('called state function name').skip(wo),
        s('(').then(wo).then(
            expr_argument_list
        ).skip(wo).skip(s(')')).optional()
    )
    yield

@generate
def return_statement():
    return (wo >> ist('return')).desc('return statement')
    yield
    
@generate
def continue_statement():
    return (wo >> ist('continue')).desc('continue statement')
    yield
    
@generate
def break_statement():
    return (wo >> ist('break')).desc('break statement')
    yield

@generate
def call_literal():
    return seq(
        regex(r'[a-zA-Z_][a-zA-Z_0-9]*').desc('called expression function name') << wo,
        (s('(') >> wo >>
            expr_argument_list
        << wo << s(')'))
    )
    yield

@generate
def label():
    return (seq(((ist("label") | ist('state')) << whitespace) >> regex(r'[a-zA-Z\_\-]+').desc('label name').tag('name'), state_body.tag('body')).map(dict) << wo)
    yield

def expression_not_empty(expr):
    if expr[0] == 'expression':
        return expression_not_empty(expr[1])
        
    elif expr[0] == 'expr':
        return len(expr[1]) > 0
        
    else:
        return bool(expr)

@generate
def templated_class_derivation():
    return seq(
        regex(r'[a-zA-Z_][a-zA-Z_0-9]*').desc('name of templated class').skip(wo),
        (
            (s('::()') >> success([])) |
            (s('::(').then(wo).then(
                parameter_list
            ).skip(wo).skip(s(')')))
        ),
        (
            wo.then(s('{')).then(
                wo.then(
                    (((ist('is') << whitespace) | string("+")) >> regex(r'[a-zA-Z0-9_\.]+').desc('flag name')).tag('flag') |
                    (((ist("isn't") << whitespace) | string('-')) >> regex(r'[a-zA-Z0-9_\.]+').desc('flag name')).tag('unflag') |

                    ist("macro").then(whitespace).then(seq(
                        regex(r'[a-zA-Z\_][a-zA-Z\_0-9]*').desc('macro name').tag('name'),
                        (wo >> s('(') >> macro_argument_list << s(')') << wo).optional().map(lambda a: a or []).tag('args'),
                        state_body.tag('body')
                    ).map(dict).tag('macro')) |
                    
                    seq(
                        (ist('set') >> whitespace).desc("'set' keyword") >> regex(r'[a-zA-Z0-9\_\.]+').tag('name'),
                        ((whitespace >> ist('to') << whitespace) | (wo >> s('=') << wo)).desc("'to' or equal sign") >> parameter.sep_by((s(',') << wo)).tag('value')
                    ).map(dict).tag('property') |

                    label.desc('override label').tag('label') |

                    ((ist('var') << whitespace) >> seq(
                        regex(r'user_[a-zA-Z0-9_]+').desc('var name').tag('name'),
                        (wo.then(s('[')).then(replaceable_number).skip(s(']'))).optional().map(lambda x: int(x or 0)).tag('size'),
                        (wo.then(s(':')).then(wo).then(regex(r'[a-zA-Z_.][a-zA-Z0-9_]+'))).desc('var type').optional().map(lambda t: t or 'int').tag('type'),
                        (wo.then(s('=')).then(expression.tag('val') | array_literal.tag('arr'))).optional().tag('value')
                    ).map(dict).tag('user var')) |

                    ((ist('array') << whitespace) >> seq(
                        regex(r'user_[a-zA-Z0-9_]+').desc('array name').tag('name'),
                        (wo >> s('=') >> wo >> array_literal).desc('array values').tag('value')
                    ).map(dict).desc('override array').tag('array')) |
                    
                    mod_block.tag('mod')
                ).skip(wo).skip(s(';')).skip(wo).many().optional()
            ).skip(s('}'))
        ).optional().map(lambda x: x or [])
    ).tag('template derivation')
    yield

@generate
def static_template_derivation():
    return wo >> (
        # anonymous derive
        seq(
            (ist('derive') >> success(None) << whitespace).tag('classname'),
            ((ist('group') << whitespace) >> group_name << whitespace).optional().map(lambda x: x or None).tag('group'),
            ist('a') >> whitespace >> templated_class_derivation.tag('source'),
        ) |

        # named derive
        seq(
            ist('derive') >> whitespace >> (regex(r'[a-zA-Z_][a-zA-Z_0-9]*').desc('name of derived class')).tag('classname') << whitespace,
            ((ist('group') << whitespace) >> group_name << whitespace).optional().map(lambda x: x or None).tag('group'),
            ist('as') >> whitespace >> templated_class_derivation.tag('source'),
        )
        
    ) << s(';') << wo
    yield

@generate
def mod_name():
    return regex(r'[a-zA-Z_-][a-zA-Z_0-9-]+')
    yield

@generate
def mod_block():
    return (seq((ist("mod") << whitespace) >> mod_name.desc('modifier name'), mod_block_body) << wo)
    yield

@generate
def class_body():
    return wo.then(
            seq(
                ist("macro") >> whitespace >> regex(r'[a-zA-Z_][a-zA-Z_0-9]*').desc('macro name').tag('name'),
                (wo >> s('(') >> macro_argument_list << s(')') << wo).optional().map(lambda a: a or []).tag('args'),
                state_body.tag('body')
            ).map(dict).tag('macro') |
            seq(
                (ist('set') >> whitespace).desc("'set' keyword") >> regex(r'[a-zA-Z0-9_\.]+').tag('name'),
                ((whitespace >> ist('to') << whitespace) | (wo >> s('=') << wo)).desc("'to' or equal sign") >> parameter.sep_by((s(',') << wo)).tag('value')
            ).map(dict).tag('property') |
            (((ist('is') << whitespace) | string("+")) >> regex(r'[a-zA-Z0-9_\.]+').desc('flag name')).tag('flag') |
            (ist('var') << whitespace) >> seq(
                regex(r'user_[a-zA-Z0-9_]+').desc('var name').tag('name'),
                (wo >> s('[') >> replaceable_number << s(']')).desc('array size').optional().map(lambda x: int(x or 0)).tag('size'),
                (wo >> s(':') >> wo >> regex(r'[a-zA-Z_.][a-zA-Z0-9_]+')).desc('var type').optional().map(lambda t: t or 'int').tag('type'),
                (wo >> s('=') >> (expression.tag('val') | array_literal.tag('arr'))).optional().tag('value')
            ).map(dict).tag('user var') |
            (((ist("isn't") << whitespace) | string('-')) >> regex(r'[a-zA-Z0-9_\.]+').desc('flag name')).tag('unflag') |
            (ist('combo') >> whitespace >> regex(r'[a-zA-Z0-9_]+').desc('combo name')).tag('flag combo') |
            seq((ist("function ") | ist('method ')) >> regex(r'[a-zA-Z_][a-zA-Z_0-9]*').desc('function name').tag('name'), state_body.tag('body')).map(dict).tag('function') |
            label.tag('label') |
            mod_block.tag('mod') |
            global_apply.tag('apply')
    ).skip(wo) << s(';') << wo
    yield

@generate
def abstract_label_body():
    return (ist("abstract label") | ist('abstract state')) >> whitespace >> regex(r'[a-zA-Z_]+').desc('label name') << s(';')
    yield

@generate
def abstract_array_body():
    return seq(
        ist("abstract array") >> whitespace >> regex(r'user_[a-zA-Z0-9_]+').desc('array name').tag('name'),
        (wo >> s('[') >> replaceable_number << s(']')).optional().map(lambda x: int(x) if x else 'any').tag('size'),
        (wo >> s(':') >> wo >> regex(r'[a-zA-Z_.][a-zA-Z0-9_]+')).desc('var type').optional().map(lambda t: t or 'int').tag('type')
    ).map(dict) << wo << s(';') << wo
    yield

@generate
def abstract_macro_body():
    return seq(
        ist("abstract macro") >> whitespace >> regex(r'[a-zA-Z_]+').desc('macro name').tag('name') << wo,
        (s('(') >> wo >> macro_argument_list << wo << s(')')).optional().map(lambda x: x or []).tag('args')
    ).map(dict) << wo << s(';') << wo
    yield

@generate
def sprite_name():
    return (
        (regex(r'[A-Z0-9_]{4}') | s('"#"')).tag('normal') |
        (ist('param') >> whitespace >> regex(r'[a-zA-Z_][a-zA-Z_0-9]*')).tag('parametrized')
    )
    yield

@generate
def superclass():
    return (
        templated_class_derivation |
        regex(r'[a-zA-Z][a-zA-Z0-9_]+').tag('classname')
    )
    yield

@generate
def group_declaration():
    return seq(
        ((ist('group') << whitespace).desc('group statement') >> wo >> group_name).tag('name'),
        (wo >> s('{') >> regex(r'[a-zA-Z0-9_]+').sep_by(s(',') << wo) << s('}')).optional().map(lambda x: x if x and tuple(x) != ('',) else []).tag('items')
    ) << s(';')
    yield

@generate
def actor_class():
    return seq(
        ((ist('actor') | ist('class')) << whitespace).desc("class statement") >> regex(r'[a-zA-Z0-9_]+').desc('class name').tag('classname'),
        ((whitespace >> ist('group') << whitespace).desc('group keyword') >> group_name).optional().map(lambda x: x or None).tag('group'),
        ((whitespace >> (ist('inherits') | ist('extends') | ist('expands'))) >> whitespace >> superclass.desc('inherited class')).optional().tag('inheritance').desc('inherited class name'),
        (whitespace >> (ist('replaces') >> whitespace >> regex(r'[a-zA-Z0-9_]+'))).desc('replaced class name').optional().tag('replacement').desc('replacement'),
        (whitespace >> s('#') >> regex(r'[0-9]+')).desc('class number').map(int).optional().tag('class number').desc('class number').skip(wo),

        ((s('{') >> wo >> class_body.many().optional() << wo.then(s('}')).skip(wo)) | s(';').map(lambda _: [])).tag('body')
    )
    yield

@generate
def templated_actor_class():
    return seq(
        (ist('actor') | ist('class') | ist('template')).desc("class template") >> s('<') >> template_parameter_list.tag('parameters') << s('>') << wo,
        regex(r'[a-zA-Z0-9_]+').desc('class name').tag('classname'),
        ((whitespace >> ist('group') << whitespace).desc('group keyword') >> group_name).optional().map(lambda x: x or None).tag('group'),
        ((whitespace >> (ist('inherits') | ist('extends') | ist('expands'))) >> whitespace >> superclass.desc('inherited class')).optional().tag('inheritance').desc('inherited class name'),
        (whitespace >> (ist('replaces') >> whitespace >> regex(r'[a-zA-Z0-9_]+'))).desc('replaced class name').optional().tag('replacement').desc('replacement'),
        (whitespace >> s('#') >> regex(r'[0-9]+')).desc('class number').map(int).optional().tag('class number').desc('class number').skip(wo),

        (s('{') >> wo >> (abstract_macro_body.desc('abstract macro').tag('abstract macro') | abstract_array_body.desc('abstract array').tag('abstract array') | abstract_label_body.desc('abstract label').tag('abstract label') | class_body).many().optional() << wo.then(s('}')).skip(wo)).tag('body')
    )
    yield

@generate
def anonymous_class():
    return seq(
        (ist('actor') | ist('class')).desc("class statement") >> \
                whitespace >> \
                (
                    (ist('inherits') | ist('extends') | ist('expands')).desc('inheritance declaration') >> \
                    whitespace >> \
                    superclass
                ).desc('inherited class name').optional().tag('inheritance') <<
                wo,
                
        ((ist('group') << whitespace).desc('group keyword') >> group_name).optional().map(lambda x: x or None).tag('group'),

        (
            ist('{') >>
            wo >>
            class_body.many().optional() <<
            wo <<
            ist('}') <<
            wo
        ).optional().map(lambda x: x or []).tag('body')
    ).tag('anonymous class')
    yield

@generate
def normal_state():
    return seq(
        sprite_name.desc('state name').skip(wo),
        (regex(r"[A-Z_.]").many() | s('"#"')).desc('state sprite').skip(wo),
        regex(r"\-?\d+").map(int).desc('state duration').skip(wo),
        modifier.many().desc('modifier').skip(wo).optional(),
        state_action.optional()
    ) | seq(
        ist('keepst').desc('\'keepst\' state').skip(whitespace) >>\
        regex(r"\-?\d+").map(int).desc('state duration').skip(wo),
        modifier.many().desc('modifier').skip(wo).optional(),
        state_action.optional()
    ).map(lambda l: [('normal', '"####"'), '"#"', *l]) | seq(
        ist('invisi').desc('\'invisi\' state').skip(whitespace) >>\
        regex(r"\-?\d+").map(int).desc('state duration').optional().map(lambda x: x or 0).skip(wo),
        modifier.many().desc('modifier').skip(wo).optional(),
        state_action.optional()
    ).map(lambda l: [('normal', 'TNT1'), 'A', *l])
    yield

@generate
def state_action():
    return action_body_repeat.tag('repeated inline body') | state_call.tag('action') | action_body.tag('inline body')
    yield

@generate
def action_body():
    return s('{') >> wo >> (state_action << s(';') << wo).many().optional() << s('}')
    yield

@generate
def replaceable_number():
    return (
        regex(r'\-?\d+') |
        variable_name
    )
    yield

@generate
def action_body_repeat():
    return seq(
        ist('x') >> wo >> replaceable_number.desc('amount of times to repeat'),
        (whitespace >> ist('index') >> whitespace >> variable_name.desc('repeat index name') << whitespace).optional().map(lambda x: x or None),
        wo >> state_action
    )
    yield

@generate
def flow_control():
    return (ist('stop') | ist('wait') | ist('fail') | ist('loop') | ist('goto') + whitespace.map(lambda _: ' ') + regex(r'[a-zA-Z0-9\_\-\:\.]+\s?(?:\+\d+)?'))
    yield

@generate
def macro_call():
    return seq(
        ( ist('from').desc("'from' determiner").then(whitespace).then(regex(r'\@*[a-zA-Z_][a-zA-Z_0-9]*').desc('extern macro classname')).skip(whitespace) ).optional().map(lambda x: x or None),
        ist('inject').desc("'inject' statement").then(whitespace).then(regex(r'\@*[a-zA-Z_][a-zA-Z_0-9]*').desc('injected macro name')).skip(wo),
        (s('(') >> expr_argument_list.desc("macro arguments") << s(')')).optional().map(lambda x: x or [])
    )
    yield

@generate
def state():
    return state_no_colon << s(';')
    yield

@generate
def state_no_colon():
    return (
        return_statement.tag('return') |
        break_statement.tag('break') |
        continue_statement.tag('continue') |
        if_statement.tag('if') |
        ifjump_statement.tag('ifjump') |
        whilejump_statement.tag('whilejump') |
        for_statement.tag('for') |
        sometimes_statement.tag('sometimes') |
        while_statement.tag('while') |
        actor_function_call.tag('call') |
        apply_block.tag('apply') |
        macro_call.tag('inject') |
        flow_control.tag('flow') |
        normal_state.tag('frames') |
        repeat_statement.tag('repeat')
    )
    yield

# Modifier effect functions.
def mod_flag(mod):
    def _eff(code, ctx, state):
        state = state.clone()
        flname = code._parse_state_modifier(ctx, mod)
    
        if all(x.upper() != flname.upper() for x in state.keywords):
            state.keywords.append(flname)
            
        yield state
    return _eff

def mod_delflag(mod):
    def _eff(code, ctx, state):
        state = state.clone()
        flname = code._parse_state_modifier(ctx, mod)
        
        while any(x.upper() == flname.upper() for x in state.keywords):
            i = next(i for i, x in enumerate(state.keywords) if x.upper() == flname.upper())
            state.keywords.pop(i)

        yield state
    return _eff

def mod_prefix(pre):
    def _eff(code, ctx, state):
        yield from code._parse_state_expr(ctx, pre)
        yield state
    return _eff

def mod_suffix(pre):
    def _eff(code, ctx, state):
        yield state
        yield from code._parse_state_expr(ctx, pre)
    return _eff

def mod_manipulate(pre):
    state_macro_name, state_body = pre

    def _eff(code, ctx, state):
        fake_macro = [state.clone()]
        
        new_ctx = ctx.derive('effect manipulation')
        new_ctx.macros[state_macro_name.upper()] = ([], fake_macro)
        
        yield from code._parse_state_expr(new_ctx, state_body)
    return _eff

@generate
def modifier_effect():
    # A modifier effect.
    return (
        (ist('+flag').desc('+flag effect') >> whitespace >> state_modifier_name).map(mod_flag) |
        (ist('-flag').desc('-flag effect') >> whitespace >> state_modifier_name).map(mod_delflag) |
        (ist('prefix').desc('prefix effect') >> whitespace >> state_body).map(mod_prefix) |
        (ist('suffix').desc('suffix effect') >> whitespace >> state_body).map(mod_suffix) |
        (ist('manipulate').desc('manipulate effect') >> whitespace >> seq(
            regex(r'[a-zA-Z_][a-zA-Z_0-9]*').desc('virtual macro name') << wo,
            state_body.desc('manipulated state body template')
        )).map(mod_manipulate)
    )
    yield

def selector_flag(name):
    def _sel(code, ctx, state):
        if not (hasattr(state, 'keywords') and state.keywords):
            return False
    
        flname = code._parse_state_modifier(ctx, name).upper()
        
        return any(flname == x.upper() for x in state.keywords)
    return _sel
    
def selector_name(name):
    def _sel(code, ctx, state):
        return hasattr(state, 'sprite') and code._parse_state_sprite(ctx, name) == state.sprite
    return _sel

@generate
def modifier_selector_basic():
    # A basic state selector, the building blocks of a
    # state selector in a modifier

    return (
        (ist('flag') >> wo >> s('(') >> state_modifier_name << s(')')).map(selector_flag) |
        (ist('sprite') >> wo >> s('(') >> sprite_name << s(')')).map(selector_name)
    )
    yield

@generate
def modifier_selector_expr():
    # A selector is a sort of selection condition expression,
    # where basic selectors are joined by boolean logic.

    return wo.then(
        modifier_selector_basic |
        ist('any').map(lambda _: (lambda _1, _2, _3: True)).desc('any selector') |

        (ist('!').desc('not operator') >> wo >> modifier_selector_expr.skip(wo)).map(lambda a: (lambda code, ctx, state: not a(code, ctx, state))) |
        (s('(') >> (
            (
                (seq(modifier_selector_expr.skip(wo >> ist('&&').desc('and operator')), wo >> modifier_selector_expr.skip(wo))).map(lambda a: (lambda code, ctx, state: a[0](code, ctx, state) and a[1](code, ctx, state))) |
                (seq(modifier_selector_expr.skip(wo >> ist('||').desc('or operator')), wo >> modifier_selector_expr.skip(wo))).map(lambda a: (lambda code, ctx, state: a[0](code, ctx, state)  or a[1](code, ctx, state))) |
                (seq(modifier_selector_expr.skip(wo >> ist('^^').desc('xor operator')), wo >> modifier_selector_expr.skip(wo))).map(lambda a: (lambda code, ctx, state: a[0](code, ctx, state)  != a[1](code, ctx, state)))
            ) |
        
            modifier_selector_expr.skip(wo)
        ) << s(')'))
    )
    yield

@generate
def modifier_clause():
    # One clause, a selector and its respective effects.
    return seq(
        wo >> modifier_selector_expr.skip(wo),
        (
            modifier_effect.map(lambda e: [e]) |
            (s('{') >> wo >> (modifier_effect << wo << s(';') << wo).at_least(1) << s('}'))
        ) << wo
    )
    yield

@generate
def mod_block_body():
    return modifier_clause.map(lambda a: [a]) | (
        wo >> s('{') >> wo >> (
            (modifier_clause << wo << s(';')).many()
        ) << wo << s('}')
    )
    yield

@generate
def state_body():
    return (
        state_no_colon.map(lambda x: [x]) | (
            wo >> string("{") >> wo >> (state).sep_by(wo) << wo << string("}")
        )
    )
    yield

@generate
def sometimes_statement():
    return seq(
        s('sometimes') >>\
                whitespace >>\
                (s('(') >> wo >> expression << wo << s(')') | replaceable_number.map(lambda r: ('expr', [('literal', (('number' if type(r) is int else 'actor variable'), r))]))).tag('chance') <<\
                s('%').optional() <<
                wo,
        state_body.optional().map(lambda x: x if x is not None else []).tag('body')
    )
    yield

@generate
def apply_block():
    return seq(
        ist('apply').desc('apply statement')
        .then(whitespace)
            .then(mod_name)
        .skip(wo),

        state_body.desc('apply block body')
        .skip(wo)
    )
    yield

@generate
def global_apply():
    return (
        ist('apply').desc('class-scoped apply statement')
            .then(whitespace)
                .then(mod_block_body.tag('body') | mod_name.tag('name'))
            .skip(wo)
    )
    yield

@generate
def if_statement():
    return seq(
        ist("if").desc('if statement')
        .then(wo)
        .then(string("("))
        .then(wo)
            .then(expression)
        .skip(wo)
        .skip(string(")"))
        .skip(wo),

        state_body.optional().map(lambda x: x if x is not None else [])
        .skip(wo),

        s(';')
        .then(wo)
        .then(s('else'))
        .then(wo)
            .then(state_body.optional().map(lambda x: x if x is not None else []))
        .skip(wo)
            .optional()
    )
    yield

def for_template(bodytype):
    return seq(
        ist("for").desc('for statement')
        .then(whitespace)
        .then(variable_name.desc('iteration parameter name'))
        .skip(whitespace),
        
        (
            ist('index').desc('index keyword')
            .then(whitespace)
            .then(variable_name.desc('iteration index name'))
            .skip(whitespace)
        ).optional().map(lambda x: x or None),

        # state_body.optional().map(lambda x: x if x is not None else [])
        ist("in")
        .then(whitespace)
        .then((
            # Possible iteration modes
            (ist("group") >> whitespace >> p_group_name << wo).tag('group') |
            (ist("range") >> whitespace >> p_range_vals << wo).tag('range')
        ))
        .skip(wo),
        
        bodytype
        .skip(wo),

        wo.then(
            (
                ist('else')
                .then(wo)
                    .then(bodytype)
                .skip(wo)
            ).optional()
        )
    )

@generate
def for_statement():
    return for_template(state_body.optional().map(lambda x: x if x is not None else []))
    yield

@generate
def static_for_loop():
    return for_template(nested_source_code.optional().map(lambda x: x if x is not None else []))
    yield

@generate
def ifjump_statement():
    return seq(
        ist("ifjump").desc('ifjump statement')
        .then(whitespace)
        .then(state_call)
        .skip(wo),

        state_body.optional().map(lambda x: x if x is not None else [])
        .skip(wo),

        s(';')
        .then(wo)
        .then(s('else'))
        .then(wo)
            .then(state_body.optional().map(lambda x: x if x is not None else []))
        .skip(wo)
            .optional()
    )
    yield

@generate
def whilejump_statement():
    return seq(
        ist("whilejump").desc('whilejump statement')
        .then(whitespace)
        .then(state_call)
        .skip(wo),

        state_body.optional().map(lambda x: x if x is not None else [])
        .skip(wo),

        s(';')
        .then(wo)
        .then(s('else'))
        .then(wo)
            .then(state_body.optional().map(lambda x: x if x is not None else []))
        .skip(wo)
            .optional()
    )
    yield

@generate
def while_statement():
    return seq(
        ist("while").desc('while statement')
            .then(wo)
            .then(string("("))
            .then(wo)
            .then(expression)
        .skip(wo)
        .skip(string(")"))
        .skip(wo),
        
        state_body.optional().map(lambda x: x if x is not None else [])
        .skip(wo),
        
        s(';')
        .then(wo)
        .then(s('else'))
        .then(wo)
            .then(state_body.optional().map(lambda x: x if x is not None else []))
        .skip(wo)
            .optional()
    )
    yield

@generate
def repeat_statement():
    return seq(
        ist('x') >> wo >> replaceable_number.desc('amount of times to repeat'),
        (whitespace >> ist('index') >> whitespace >> variable_name.desc('repeat index name') << whitespace).optional().map(lambda x: x or None),
        wo >> state_body
    )
    yield

@generate
def anonymous_macro():
    return seq(
        ist("macro") >> (s('(') >> macro_argument_list << s(')') << wo).optional().map(lambda a: a or []),
        state_body
    ).tag('anonymous macro')
    yield

@generate
def macro_def():
    return seq(
        ist("macro") >> whitespace >> regex(r'[a-zA-Z_][a-zA-Z_0-9]*').desc('macro name').tag('name'),
        (wo >> s('(') >> macro_argument_list << s(')') << wo).optional().map(lambda a: a or []).tag('args'),
        state_body.tag('body')
    ).map(dict).tag('macro')
    yield

@generate
def nested_source_code():
    return ist('{') >> source_code << ist('}') << wo << ist(';')
    yield

@generate
def source_code():
    return wo.desc('ignored whitespace') >> (group_declaration.tag('group') | macro_def.tag('macro') | actor_class.tag('class') | static_template_derivation.tag('static template derivation') | templated_actor_class.tag('class template') | static_for_loop.tag('for')).sep_by(wo) << wo.desc('ignored whitespace')
    yield

@generate
def source_code_top():
    return source_code
    yield

def rpcont(i=0):
    @generate
    def recursive_paren_content():
        return ((regex(r'[^\(\)]') if i else regex(r'[^\(\)\,]')) | (s('(') + rpcont(i+1)) + s(')')).many().concat()
        yield

    return recursive_paren_content

def preprocess_for_macros(code, defines=(), this_fname=None, i=0):
    l = code

    for key, (val, d_args) in defines.items():
        nl = ""

        while True:
            j = re.search(r'\b' + re.escape(key.upper()) + r'\(', l.upper())

            if not j:
                nl += l
                break

            j = j.start()
            nl += l[:j]
            l = l[j + len(key):]

            try:
                args, l = (s('(') >> rpcont().sep_by(s(',') << wo) << s(')')).parse_partial(l)
                args = [x for x in args if x]

            except parsy.ParseError as e:
                print('\n')
                traceback.print_exc()
                print()

                raise PreprocessingError("Unexpected parse error using parametrized preprocessor alias '{}' ({})".format(key, e), i, this_fname, code + '\n\n')

            if len(d_args) != len(args):
                raise PreprocessingError("Bad number of arguments using parametrized preprocessor alias '{}': expected {}, got {}!".format(key, len(d_args), len(args)), i, this_fname, code)

            v = val

            for aname, aval in zip(d_args, args):
                #ov = v
                v = re.sub(r'\({}\)'.format(re.escape(aname)), '({})'.format(aval), v)

            res = preprocess_for_macros(v, defines, this_fname, i)
            nl += res

        l = nl

    return l

def preprocess_code(code, imports=(), defs=(), defines=(), this_fname=None, rel_dir='.', begin_char=None):
    if not begin_char:
        begin_char = [0]

    if not isinstance(imports, dict):
        imports = dict(imports)

    if not isinstance(defs, dict):
        defs = dict(defs)

    if not isinstance(defines, dict):
        defines = dict(defines)

    # conditional substitution
    pcodelines = []

    cond_active = True
    depth = 0
    check_depth = 0

    # comments
    src_lines = code.split('\n')

    code = re.sub(r'\/\/[^\r\n]+', lambda x: re.sub(r'[^\n]', ' ', x.group(0)), code)
    code = re.sub(r'\/\*(\n|.)+?\*\/', lambda x: re.sub(r'[^\n]', ' ', x.group(0)), code)
    code = re.sub(r'\\\n', '', code)

    codelines = code.split('\n')

    for _i, l in enumerate(codelines):
        i = _i + 1
        src_l = src_lines[_i]

        check_line_case = re.sub(r'\s+', ' ', re.sub(r'^\s+', '', l))
        check_line = check_line_case.upper()

        # conditional blocks
        if check_line.startswith('#IFDEF ') or check_line.startswith('#IFDEFINED '):
            check_depth += 1

            if cond_active:
                key = check_line_case.split(' ')[1]

                if key not in defs:
                    cond_active = False

            else:
                depth += 1

        elif ifundef.match(check_line.split(' ')[0]):
            check_depth += 1

            if cond_active:
                key = check_line_case.split(' ')[1]

                if key in defs:
                    cond_active = False

            else:
                depth += 1

        elif check_line.startswith('#IFEQ ') or check_line.startswith('#IFEQUAL ') or check_line.startswith('#IFEQUALS '):
            check_depth += 1

            if cond_active:
                key = check_line_case.split(' ')[1]
                value = check_line_case.split(' ')[2:]

                if key not in defs or defs[key] != value:
                    cond_active = False

            else:
                depth += 1

        elif ifneq.match(check_line.split(' ')[0]):
            check_depth += 1

            if cond_active:
                key = check_line_case.split(' ')[1]
                value = check_line_case.split(' ')[2:]

                if key not in defs or defs[key] == value:
                    cond_active = False

            else:
                depth += 1

        elif check_line.startswith('#ELSE') or check_line.startswith('#OTHERWISE'):
            if check_depth == 0:
                raise PreprocessingError("Attempted to use 'else' on a conditional preprocessor block that didn't exist!", i, this_fname, check_line_case)

            cond_active = not cond_active

        elif check_line.startswith('#ENDIF'):
            check_depth -= 1

            if check_depth < 0:
                raise PreprocessingError("Attempted to end a conditional preprocessor block that didn't exist!", i, this_fname, check_line_case)

            if not cond_active:
                if depth > 0:
                    depth -= 1

                else:
                    cond_active = True

        # imports and definitions
        if cond_active:
            if check_line.startswith('#IMPORT ') or check_line.startswith('#INCLUDE '):
                gfname = os.path.join(rel_dir, ' '.join(check_line_case.split(' ')[1:]))
                gname = glob.glob(gfname)
                
                if gname:
                    for fname in gname:
                        if not os.path.isfile(fname):
                            raise PreprocessingError("The module '{}' was not found".format(fname), i, this_fname, check_line_case)

                        if imports.get(fname, object()) == this_fname:
                            raise PreprocessingError("The module '{}' was found in an infinite import cycle!".format(fname), i, this_fname, check_line_case)

                        elif fname not in imports:
                            imports[fname] = this_fname

                            with open(fname) as fp:
                                lines = preprocess_code(fp.read() + "\n", imports, defs, defines, this_fname=fname, rel_dir=os.path.dirname(fname), begin_char=begin_char)

                                for l in lines:
                                    pcodelines.append(l)
                            
                else:
                    raise PreprocessingError("No module matching '{}' was found".format(gfname), i, this_fname, check_line_case)
                        
            elif check_line.startswith('#RESOURCE ') or check_line.startswith('#RES '):
                # Add resource to output PK3
                key = check_line_case.split(' ')[1]
                value = ' '.join(check_line_case.split(' ')[2:])

                if value == '':
                    defs[key] = None

                else:
                    defs[key] = value

            elif check_line.startswith('#DEF ') or check_line.startswith('#DEFINE '):
                key = check_line_case.split(' ')[1]
                value = ' '.join(check_line_case.split(' ')[2:])

                if value == '':
                    defs[key] = None

                else:
                    defs[key] = value

            elif defmacro.match(check_line.split(' ')[0]):
                name = check_line_case.split(' ')[1]
                value = ' '.join(check_line_case.split(' ')[2:])

                (key, args), _ = seq(regex(r'[a-zA-Z_][a-zA-Z0-9_]*'), (s('(') >> (regex(r'[a-zA-Z_][a-zA-Z0-9_]*').sep_by(s(','))) << s(')')).optional().map(lambda x: x or [])).parse_partial(name)
                args = [x for x in args if x]

                if value == '':
                    raise PreprocessingError("Empty preprocessor macros ('{}') are not allowed!".format(key), i, this_fname, check_line_case)

                defs[key] = value
                defines[key] = (value, args)

            elif check_line.startswith('#UNDEF ') or check_line.startswith('#UNDEFINE '):
                key = check_line_case.split(' ')[1]

                if key in defs:
                    defs.pop(key)

                if key in defines:
                    defines.pop(key)

        if not check_line.startswith('#') and cond_active:
            l = preprocess_for_macros(l, defines, this_fname, i)
            src_l = preprocess_for_macros(src_l, defines, this_fname, i)

            pcodelines.append((this_fname, i, src_l, l, begin_char[0]))
            begin_char[0] += len(l) + 1

    return pcodelines

def parse_postcode(postcode, error_handler=None):
    try:
        lim = sys.getrecursionlimit()
        sys.setrecursionlimit(lim * 16)
        clazzes = source_code_top.parse('\n'.join(l[3] for l in postcode))
        sys.setrecursionlimit(lim)

        return clazzes

    except parsy.ParseError as parse_err:
        m = error_line.match(str(parse_err))

        if m is None:
            raise

        else:
            err = ZDParseError(parse_err, postcode[int(m[2])])

            if error_handler is None:
                raise err

            else:
                error_handler(err)

def parse_code(code, filename=None, dirname='.', error_handler=None, preproc_defs=(), imports=()):
    try:
        return parse_postcode(preprocess_code(code, defs=preproc_defs, this_fname=filename, rel_dir=dirname, imports=imports), error_handler=error_handler)

    except PreprocessingError as pperr:
        if error_handler is None:
            raise

        else:
            error_handler(pperr)



import argparse
import os
import sys

try:
    import zdcode
    import zdcode.zake as zake
    from zdcode.bundle import Bundle

except ImportError:
    import __init__ as zdcode
    import zake
    from bundle import Bundle


def main():
    return main_zake()


def main_zake():
    return zake.main(print_status_code=False)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zdcode']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=20.3.0,<21.0.0', 'parsy>=1.3.0,<2.0.0']

entry_points = \
{'console_scripts': ['zake = zdcode.program:main_zake',
                     'zdcode = zdcode.program:main']}

setup_kwargs = {
    'name': 'zdcode',
    'version': '2.12.2',
    'description': 'A ZDoom DECORATE transpiler and mod content build system',
    'long_description': '# ZDCode 2.0\n\n"The language that compiles to ye olde DECORATE!"\n\nZDCode is a project that aims to make writing DECORATE _better_; that is,\nto expand the possibilities not only of what the code itself can do, but\nalso of how it can be written, or concerted with other ZDCode projects\nand authors, or distributed to modders and players alike. ZDCode is\nan attempt to make modding for ZDoom-based platforms, like Zandronum,\nmuch more similar to the ecosystem of an actual language, like a C linker,\nor a JavaScript web bundler.\n\nTake this example:\n\n```c\n#UNDEFINE ANNOYING\n\nclass RunZombie inherits ZombieMan replaces ZombieMan #2055\n{\n    set Gravity to 0.4; // high up...\n    set Speed to 0;\n    is NOBLOCKMONST;\n    set Speed to 0;\n\n    label See\n    {\n        inject SeeCheck;\n        POSS AB 5 A_Recoil(-0.7);\n        inject SeeCheck;\n        POSS AB 4 A_Recoil(-0.7);\n        inject SeeCheck;\n        POSS ABCD 3 A_Recoil(-0.7);\n        inject SeeCheck;\n        goto RunLoop;\n    };\n\n    macro SeeCheck\n    {\n        TNT1 A 0 A_Chase;\n        POSS A 0 A_FaceTarget();\n    };\n\n    macro ZombieJump\n    {\n        if ( health > 5 )\n            return;\n\n        while ( z == floorz )\n        {\n            POSS A 5 [Bright];\n            POSS A 11 ThrustThingZ(0, 30, 0, 1);\n        };\n\n        #ifndef ANNOYING\n        while ( z > floorz )\n        {\n            POSS AB 4;\n        };\n\n        POSS G 9;\n        POSS B 22;\n        #endif\n\n        POSS AB 2 A_Chase;\n    };\n\n    label RunLoop\n    {\n        x 2\n        {\n            POSS ABCD 2 A_Recoil(-0.7);\n            inject SeeCheck;\n        };\n\n        inject ZombieJump;\n        loop;\n    };\n}\n```\n\nThis is what happens when that beauty goes through **ZDCode 2.0**:\n\n```\nActor _Call_NPaLK2i4Etrk1DERaszVFVbnG6JiT6KwJHX_0 : Inventory {Inventory.MaxAmount 1}\nActor _Call_NPaLK2i4Etrk1DERaszVFVbnG6JiT6KwJHX_1 : Inventory {Inventory.MaxAmount 1}\nActor _Call_NPaLK2i4Etrk1DERaszVFVbnG6JiT6KwJHX_2 : Inventory {Inventory.MaxAmount 1}\nActor _Call_NPaLK2i4Etrk1DERaszVFVbnG6JiT6KwJHX_3 : Inventory {Inventory.MaxAmount 1}\nActor _Call_NPaLK2i4Etrk1DERaszVFVbnG6JiT6KwJHX_4 : Inventory {Inventory.MaxAmount 1}\nActor _Call_NPaLK2i4Etrk1DERaszVFVbnG6JiT6KwJHX_5 : Inventory {Inventory.MaxAmount 1}\nActor _Call_NPaLK2i4Etrk1DERaszVFVbnG6JiT6KwJHX_6 : Inventory {Inventory.MaxAmount 1}\n\n\nActor RunZombie : ZombieMan replaces ZombieMan 2055\n{\n    Gravity 0.4\n    Speed 0\n    Speed 0\n\n    +NOBLOCKMONST\n\n    States {\n        F_SeeCheck:\n            TNT1 A 0 A_Chase\n            POSS A 0 A_FaceTarget\n            TNT1 A 0 A_JumpIfInventory("_Call_NPaLK2i4Etrk1DERaszVFVbnG6JiT6KwJHX_0", 1, "_CLabel0")\n            TNT1 A 0 A_JumpIfInventory("_Call_NPaLK2i4Etrk1DERaszVFVbnG6JiT6KwJHX_1", 1, "_CLabel1")\n            TNT1 A 0 A_JumpIfInventory("_Call_NPaLK2i4Etrk1DERaszVFVbnG6JiT6KwJHX_2", 1, "_CLabel2")\n            TNT1 A 0 A_JumpIfInventory("_Call_NPaLK2i4Etrk1DERaszVFVbnG6JiT6KwJHX_3", 1, "_CLabel3")\n            TNT1 A 0 A_JumpIfInventory("_Call_NPaLK2i4Etrk1DERaszVFVbnG6JiT6KwJHX_4", 1, "_CLabel4")\n            TNT1 A 0 A_JumpIfInventory("_Call_NPaLK2i4Etrk1DERaszVFVbnG6JiT6KwJHX_5", 1, "_CLabel5")\n            TNT1 A -1\n\n            F_ZombieJump:\n            TNT1 A 0 A_JumpIf(!(health > 5), 2)\n            TNT1 A 0 A_JumpIfInventory("_Call_NPaLK2i4Etrk1DERaszVFVbnG6JiT6KwJHX_6", 1, "_CLabel6")\n            Stop\n            TNT1 A 0\n        _WhileBlock0:\n            TNT1 A 0 A_JumpIf(!(z == floorz), 3)\n            POSS A 5  Bright\n            POSS A 11 ThrustThingZ(0, 30, 0, 1)\n            Goto _WhileBlock0\n            TNT1 A 0\n        _WhileBlock1:\n            TNT1 A 0 A_JumpIf(!(z > floorz), 3)\n            POSS A 4\n            POSS B 4\n            Goto _WhileBlock1\n            TNT1 A 0\n            POSS G 9\n            POSS B 22\n            POSS A 2 A_Chase\n            POSS B 2 A_Chase\n            TNT1 A 0 A_JumpIfInventory("_Call_NPaLK2i4Etrk1DERaszVFVbnG6JiT6KwJHX_6", 1, "_CLabel6")\n            TNT1 A -1\n\n        See:\n            TNT1 A 0 A_GiveInventory("_Call_NPaLK2i4Etrk1DERaszVFVbnG6JiT6KwJHX_0")\n            Goto F_SeeCheck\n        _CLabel0:\n            TNT1 A 0 A_TakeInventory("_Call_NPaLK2i4Etrk1DERaszVFVbnG6JiT6KwJHX_0")\n            POSS A 5 A_Recoil(-0.7)\n            POSS B 5 A_Recoil(-0.7)\n            TNT1 A 0 A_GiveInventory("_Call_NPaLK2i4Etrk1DERaszVFVbnG6JiT6KwJHX_1")\n            Goto F_SeeCheck\n        _CLabel1:\n            TNT1 A 0 A_TakeInventory("_Call_NPaLK2i4Etrk1DERaszVFVbnG6JiT6KwJHX_1")\n            POSS A 4 A_Recoil(-0.7)\n            POSS B 4 A_Recoil(-0.7)\n            TNT1 A 0 A_GiveInventory("_Call_NPaLK2i4Etrk1DERaszVFVbnG6JiT6KwJHX_2")\n            Goto F_SeeCheck\n        _CLabel2:\n            TNT1 A 0 A_TakeInventory("_Call_NPaLK2i4Etrk1DERaszVFVbnG6JiT6KwJHX_2")\n            POSS A 3 A_Recoil(-0.7)\n            POSS B 3 A_Recoil(-0.7)\n            POSS C 3 A_Recoil(-0.7)\n            POSS D 3 A_Recoil(-0.7)\n            TNT1 A 0 A_GiveInventory("_Call_NPaLK2i4Etrk1DERaszVFVbnG6JiT6KwJHX_3")\n            Goto F_SeeCheck\n        _CLabel3:\n            TNT1 A 0 A_TakeInventory("_Call_NPaLK2i4Etrk1DERaszVFVbnG6JiT6KwJHX_3")\n            goto RunLoop\n\n        RunLoop:\n            POSS A 2 A_Recoil(-0.7)\n            POSS B 2 A_Recoil(-0.7)\n            POSS C 2 A_Recoil(-0.7)\n            POSS D 2 A_Recoil(-0.7)\n            TNT1 A 0 A_GiveInventory("_Call_NPaLK2i4Etrk1DERaszVFVbnG6JiT6KwJHX_4")\n            Goto F_SeeCheck\n        _CLabel4:\n            TNT1 A 0 A_TakeInventory("_Call_NPaLK2i4Etrk1DERaszVFVbnG6JiT6KwJHX_4")\n            POSS A 2 A_Recoil(-0.7)\n            POSS B 2 A_Recoil(-0.7)\n            POSS C 2 A_Recoil(-0.7)\n            POSS D 2 A_Recoil(-0.7)\n            TNT1 A 0 A_GiveInventory("_Call_NPaLK2i4Etrk1DERaszVFVbnG6JiT6KwJHX_5")\n            Goto F_SeeCheck\n        _CLabel5:\n            TNT1 A 0 A_TakeInventory("_Call_NPaLK2i4Etrk1DERaszVFVbnG6JiT6KwJHX_5")\n            TNT1 A 0 A_GiveInventory("_Call_NPaLK2i4Etrk1DERaszVFVbnG6JiT6KwJHX_6")\n            Goto F_ZombieJump\n        _CLabel6:\n            TNT1 A 0 A_TakeInventory("_Call_NPaLK2i4Etrk1DERaszVFVbnG6JiT6KwJHX_6")\n            Goto RunLoop\n    }\n}\n```\n\nWhile the compile output does look unreadable, expecting readability from it is akin to expecting\nreadability from a binary executable file. This is the principle of using ZDCode instead of\nDECORATE, after all – a more readable, concise, and organized, way to write DECORATE.\n\nJust slap the output in your WAD and... [look at what happens!](https://i.imgur.com/mr5wJ85.gifv)\n\n\n\n# Design Concepts\n\n\n\n### Bundling\n\nSimilar to web bundling technologies, such as Browserify, ZDCode tackles the issue of incompatibilities\namong different generated DECORATE files by instead focusing on _bundling_ the ZDCode input into a single\nDECORATE file.\n\nThe compiled output of ZDCode, much like a compiled C program after being linked, is treated as an\nindivisible, integral, and immutable chunk of instructions, to be read and interpreted exclusively by\nZDoom. Instead of attempting to merge separate DECORATE outputs, it is much easier, and in fact more\nplausible, to link other ZDCode projects, libraries, and source code in general, akin to libraries\nin interpreted languages, like Python. This is also how web bundling technologies operate.\n\n\n### Non-[Imperativity](https://en.wikipedia.org/wiki/Imperative_programming) and other lower-level limitations\n\nUnlike other programming languages, ZDCode has a very specific purpose.\n\nC is a portable way of writing a relatively abstract representation of machine instructions (e.g.\n`myNum += 1` instead of `ADD EAX, 1`), and interpreted languages are instructions for an interpreter\'s\nvirtual machine, whose computations do still, in the end, reflect machine code.\n\nZDCode does not have the same capabilities of using any arbitrary computer resources, because DECORATE\nitself doesn\'t. Rather, DECORATE is a mere form of expressing "classes" (which are only templates for\nobjects in ZDoom-based source ports, rather than actual object-oriented programming constructs), which\nin turn guide "actors" (fancy name for game objects, instead of being strictly actor model constructs),\nincluding their actions, properties, and states (frames).\n\nFor this reason, ZDCode is not concerned with concepts of actual imperative programming, like actual `for`\nloops that actually increment a discrete integer variable towards an arbitrary limit. Rather, it tries\nto make it easier, simpler, more programming-friendly and more systematic to write DECORATE behaviour for\nZDoom, without relying on ZScript and requiring modern GZDoom versions for mere language support\nreasons. Zandronum does not support ZScript.\n\nDECORATE itself does not support sharing behaviour between separate actor classes; rather, it supports\nusing actor classes from other classes, but only in indirect ways, such as spawning other actors. It\nhas very limited interaction among different actors, which are more centered around basic concepts that\nevery physical actor has (such as collision, movement, and to an extent AI behaviour), without allowing\nactual actor-model-esque messages to be passed directly between those game objects. At this point,\nwriting DECORATE in any way that attempts to concert multiple objects becomes contrived and relies on\nmostly unrelated behaviour intended for very different things; trying to send a message between two\nactors is similar to using a bottle, usually a small liquid container, as a means of propagating a\nmessage across a river, or a lake. It does work, but it relies on the buoyancy and imperviousness\nof the bottle. It\'s much better to use something designed for this, like in this example, it could be\nradio waves, or an actual wire.\n\nUnfortunately, ZDCode is not able to overcome these limitations at run-time. It still cannot have\nimp Bob tell imp Joe that the marine is coming, and that they should ambush from different hiding\nspots. What it _can_ do, however, is make the _writing_ part of the code a lot simpler, by providing\ntools to exploit behaviours.\n\n\n### Distribution\n\nThe concept of _distribution_ in ZDCode is intimately related to that of bundling, specifically\nbecause it concerns with the ease of availability of libraries to the programmer, and also the\nease of distribution and dependency management.\n\nAt one point, it was planned to create a rather standardized format for fetching ZDCode packages\nusing indexes stored in Git. However, this has been wholly deemed unnecessary. Instead, the\ncurrent roadmap is to add simple support via transports like HTTP, Git and FTP, and allowing\nother transport implementations as well, using a termina-friendly URI-like format. This is very\nreminiscent of the syntax Go package management uses, e.g. `go get github.com/someone/something`.\n\nThis deals with two issues at the same time: it ensures both that players can easily retrieve mods (and\nupdates thereof) directly from the Internet (automatically bundling them if necessary), whilst at\nthe same time enabling ZDCode mod authors to both obtain and share code more efficiently, both for libraries\nand finished mods.\n\n\n# Programming Constructs\n\nWelcome to the main attractions!\n\n\n## Code Blocking\n\nThis may seem like a primitive part of a programming language, but DECORATE uses states\nand labels, instead of code blocking. It\'s more akin to Assembly (or oldschool C), with\njumps rather than groups of statements.\n\nZDCode allows the programmer to group their state code into blocks, useful for statements\nlike repetition (`x 5 { A_Print ("I\'m said five times!") }`), control flow, or even\nmere readability.\n\n\n\n## Macros\n\nMacros are a way to inject common behaviour into different locations, as states that are\nused multiple times.\n\n```\n// Global macros!\nmacro ThrustUpward(amount) {\n    TNT1 A 0 ThrustThingZ(0, amount, 0, 1);  // My, DECORATE has some convoluted things sometimes.\n}\n\nclass YipYipHurrahMan extends ZombieMan {\n    // Class-local macros!\n    macro Yip {\n        inject ThrustUpward(20);\n    };\n\n    macro Hurrah {\n        inject ThrustUpward(90);\n    };\n\n    label Spawn {\n        POSS A 12;   // \'12\' means short yips\n          inject Yip;\n        POSS B 12;\n          inject Yip;\n        POSS D 50;   // \'50\' means long hurrahs\n          inject Hurrah;\n          loop;\n    };\n}\n```\n\nThey are simple, because the states in them are simply copied at compile time, instead of\ncalled at runtime. (Functions are legacy, unreliable, and now deprecated.) They support\nstatic parameters, as well. They can\'t change at runtime, but they do make life easier too.\n\n\n\n## Conditions\n\nIn contrast to DECORATE\'s highly manual and tediously finnicky (and almost Assembly-like)\nstate jumps, ZDCode boasts a much nicer format, that does not require offset maintenance\nin the source code, nor separate state labels, and that is easier to both integrate with\nexisting code, extend with new code, or nest with more conditions and other constructs.\n\n```\nclass RedBlue {\n    label Spawn {\n        if (z > floorz + 64) {\n            // Red sprite, normal gravity (except half, but you know).\n            RDBL R 2 A_SetGravity(0.5);\n        };\n\n        else {\n            // Blue sprite, reverse gravity.\n            RDBL B 2 A_SetGravity(-0.5);\n        };\n        \n        loop;\n    };\n}\n```\n\n\n## Preprocessor\n\nYes, there is a C-like preprocessor in ZDCode! It has the usual `#DEFINE`, `#IFDEF`,\n`#IFNDEF`, and the fundamental part of using any library - `#INCLUDE`. Among other things,\ntoo.\n\n```\nclass LocalizedZombie extends Zombieman replaces Zombieman {\n    // This is a merely illustrative example.\n    // For actual locatization, please just use ZDoom\'s LANGUAGE lump instead.\n    // Apart from that, it demonstrates the effectiveness of the otherwise\n    // simple and rudimentary preprocessor ZDCode uses.\n\n    macro SeeMessage(message) {\n        invisi A_Print(message); // invisi == TNT1, also duration defaults to 0\n\n        // Any better ideas for a message printing macro? I\'m all ears!\n    };\n\n    label See {\n        #ifdef LANG\n            #ifeq LANG EN_US\n                inject SeeMessage("Hey, you!");\n            #else\n            #ifeq LANG PT_BR\n                inject SeeMessage("Ei, você!");\n            #else // I know, not very pretty. Python gets away with it, though!\n            #ifeq LANG DE_DE\n                inject SeeMessage("Achtung! Halt!");\n            #else\n                inject SeeMessage("/!\\"); // Attention?\n            #endif\n            #endif\n            #endif\n            #endif\n        #endif\n        goto Super.See;\n    };\n}\n```\n\n\n## Templates\n\nIt was already possible to have a class inherit another. It is very simple DECORATE\nbehaviour that ZDCode of course permits as well, although with a bit cleaner syntax.\n\nIn addition to that, ZDCode allows instantiating multiple classes that differ slightly\nfrom a base _template_. The difference between this and inheritance is that, rather than\nhappening at load time (where ZDoom reads the DECORATE), it happens at compile-time, which\nmeans many cool tricks can be done by using this alongside other ZDCode features.\n\n```\nclass<size> ZombieSizes extends Zombieman {\n    set Scale to size;\n};\n\nderive SmallZombie  as ZombieSizes::(0.5);\nderive BiggerZombie as ZombieSizes::(1.3);\nderive SuperZombie  as ZombieSizes::(2.2);\n```\n\nDerivations can optionally include extra definitions, including the ability to \'implement\'\n**abstract macros**, **abstract labels** and define the values of **abstract arrays** that\nthe template may specify.\n\n```\nclass<> FiveNumbers {\n    abstract array numbers;\n\n    abstract macro PrintNumber(n);\n\n    label Spawn {\n        // for ranges are not supported yet\n        \n        inject PrintNumber(numbers[0]);\n        inject PrintNumber(numbers[1]);\n        inject PrintNumber(numbers[2]);\n        inject PrintNumber(numbers[3]);\n        inject PrintNumber(numbers[4]);\n\n        stop;\n    };\n}\n\nderive PlainFibonacci as FiveNumbers::() {\n    macro PrintNumber(n) {\n        TNT1 A 0 A_Log(n);\n    };\n\n    array numbers { 1, 1, 2, 3, 5 };\n}\n```\n\nIt is recommended that classes that derive a template also inherit a base class, even\nif for purely symbolic reasons, since it helps organize the code a bit, and so classes\nthat derive from such templates can be listed property by ZDoom\'s `dumpclasses` command\n(using the inheritance base class as an argument), among other things.\n\n\n## Groups and Group Iteration\n\nA _group_ is a compile-time sequence of literals that can be iterated using a `for in` loop.\nClass syntax allows adding the class\' name to a group. More importantly, templates can also\nspecify a group, but rather than the template itself, the names of all derived classes are\nadded to the group when the code is parsed by the ZDCode compiler.\n\n```\ngroup fruits;\n\nclass<name> Fruit group fruits {\n    macro PrintMe() {\n        TNT1 A 0 A_Print(name);\n    };\n}\n\nderive Orange as Fruit::("Orange");\nderive Banana as Fruit::("Banana");\nderive Lemon  as Fruit::("Lemon");\nderive Grapes as Fruit::("Grapes");\n\nclass FruitPrinter {\n    label Spawn {\n        // \'index ___\' is optional, but retrieves\n        // the # of the iteration, like the i in a\n        // C \'for (i = 0; i < n; i++)\' loop, and can be\n        // quite useful\n        \n        for fruitname index ind in fruits {\n            // Log the index and fruit classname, like some sort of debug example\n            \n            TNT1 A 0 A_Log("Fruit #, Fruit Name");\n            TNT1 A 0 A_Log(ind);\n            TNT1 A 0 A_Log(fruitname);\n\n            // Call the derived class macro from FruitPrinter. Yes.\n            // The \'from\' keyword means that the macro is from a\n            // different class. The at sign means that the class\n            // name is taken from the parameter \'fruitname\', rather\n            // than it being the name of the class itself. It can\n            // also be used in the macro name part for interesting\n            // tricks, similar to C\'s function pointer syntax.\n\n            from @fruitname inject PrintMe();\n        };\n\n        stop;\n    };\n}\n```\n\n\n## State Modifiers\n\nAnother powerful feature of ZDCode is the ability to modify states at runtime,\nwhere each modifier applies certain effects based on certain selectors.\n\nEach modifier is actually a list of clause, where a clause pairs a selector with\none or more effects.\n\n```\nclass DiscoZombie extends Zombieman {\n    mod DiscoLight {\n        (sprite(TNT1)) suffix TNT1 A 0 A_PlaySound("disco/yeah"); // now we can use TNT1 frames to play some weird sound!\n\n        (all) +flag Bright; // Always bright, all the time!\n    };\n\n    label Spawn {\n        apply DiscoLight {\n            POSS AB 4;\n        };\n\n        loop;\n    };\n\n    // You can also apply the modifier to other state labels (like See or Missile)\n    // using the apply syntax, but you get the gist.\n}\n```\n\n\n\n\n',
    'author': 'Gustavo6046',
    'author_email': 'rehermann6046@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
